package org.enhydra.barracuda.contrib.dbroggisch.repopulation;

import java.util.List;
import java.util.ArrayList;
import org.enhydra.barracuda.core.forms.DefaultFormElement;
import org.enhydra.barracuda.core.comp.ListModel;
import org.enhydra.barracuda.core.comp.DefaultListModel;
import org.apache.log4j.Logger;
import org.enhydra.barracuda.core.comp.ListSelectionModel;
import org.enhydra.barracuda.core.forms.FormType;
import org.enhydra.barracuda.core.forms.FormValidator;
import org.enhydra.barracuda.core.comp.ItemMap;
import org.enhydra.barracuda.core.comp.ViewContext;
import org.enhydra.barracuda.core.comp.BSelect;

/**
 * Use this FormElement to repopulate the request value
 * into a select box created by the &lt;select&gt;-tag. 
 *
 * @author <a href="mailto:deets@web.de">Diez B. Roggisch</a>
 * @version 1.0
 */
public class SelectFormElement extends DefaultFormElement implements RepopulationElement {
    private static final Logger logger = Logger.getLogger(SelectFormElement.class.getName());
    ListModel lm;
    ListSelectionModel lsm;
    int selectionType = -1;

    /**
     * Set the list of allowed values. 
     *
     * @param lm a <code>ListModel</code> value
     */
    public void setListModel(ListModel lm) {
        this.lm = lm;
    }

    /**
     * Get the list of allowed values.
     *
     * @return a <code>ListModel</code> value
     */
    public ListModel getListModel() {
        return lm;
    }

    /**
     * Set the ListSelectionModel that determines which entries
     * of the select box are rendered.
     *
     * @param lsm a <code>ListSelectionModel</code> value
     */
    public void setListSelectionModel(ListSelectionModel lsm) {
        this.lsm = lsm;
    }

    
    /**
     * Get the ListSelectionModel.
     *
     * @return a <code>ListSelectionModel</code> value
     */
    public ListSelectionModel getListSelectionModel() {
        if(logger.isDebugEnabled()) {
            logger.debug("Name: " + getName());
            logger.debug("allowMultiples: " + (
                                               allowMultiples() ? "yes" : "no"));
            if(getVals() != null) {
                logger.debug("type(Vals): " + getVals().getClass().getName());
                logger.debug("Vals: " + getVals());
            }
        }
        if(lsm == null) {
            int selType = selectionType;
            if(selType == -1)  {
                if(allowMultiples()) {
                    selType = ListSelectionModel.MULTIPLE_INTERVAL_SELECTION;
                } else {
                    selType = ListSelectionModel.SINGLE_SELECTION;
                }
            }
            List selected;
            if(getOrigVal() != null) {
                if(getOrigVal() instanceof List) {
                    selected = (List)getOrigVal();
                } else {
                    selected = new ArrayList();
                    selected.add(getOrigVal());
                }
            } else {
                // If we have Vals (set be the
                selected = new ArrayList();
            }
            if(lm instanceof DefaultListModel) {
                return calcListSelectionModel(((DefaultListModel)lm).subList(0, lm.getSize()), selected, selType);
            }
            else {
                List values = new ArrayList(lm.getSize());
                for(int i = 0; i < lm.getSize(); i++) {
                    values.add(lm.getItemAt(i));
                }
                return calcListSelectionModel(values, selected, selType);
            }
        }
        return lsm;
    }


    /**
     * This is a helper method that takes a list of permitted values, and a subset
     * of selected values out of the former list. It will then compute the mode that
     * fits this selection. See {@link org.enhydra.barracuda.core.comp.ListSelectionModel ListSelectionModel}.
     * 
     *
     * @param vals a <code>List</code> of permitted values
     * @param sel a <code>List</code> which specifies a subset of <b>vals</b>
     * @return the mode, as specified in {@link org.enhydra.barracuda.core.comp.ListSelectionModel ListSelectionModel}
     */
    public static int calcSelectionMode(List vals, List sel) {
        ListSelectionModel lsm = calcListSelectionModel(vals, sel, ListSelectionModel.MULTIPLE_INTERVAL_SELECTION);
        return calcSelectionMode(lsm);
    }

    /**
     * This method computes the actual mode for the given ListSelectionModel.
     * See {@link org.enhydra.barracuda.core.comp.ListSelectionModel ListSelectionModel} for the range
     * of permitted values.
     * 
     * @param lsm a <code>ListSelectionModel</code> value
     * @return an <code>int</code> value
     */
    public static int calcSelectionMode(ListSelectionModel lsm) {
        int selMode = -1;
        int min = lsm.getMinSelectionIndex();
        if(min != -1) {
            int max = lsm.getMaxSelectionIndex();
            if(min != max) {
                boolean singleInterval = true;
                for(int i = min; i <= max; i++) {
                    singleInterval = singleInterval && lsm.isSelectedIndex(i);
                }
                if(singleInterval) {
                    selMode = ListSelectionModel.SINGLE_INTERVAL_SELECTION;
                } else {
                    selMode = ListSelectionModel.MULTIPLE_INTERVAL_SELECTION;
                }
            } else {
                selMode = ListSelectionModel.SINGLE_SELECTION;
            }
        }
        return selMode;
    }

    /**
     * A convenience method that takes two lists, the latter beeing a subset of the first, and
     * computes the ListSelectionModel for them.
     *
     * @param vals a <code>List</code> value
     * @param sel a <code>List</code> value
     * @param selType see {@link org.enhydra.barracuda.core.comp.ListSelectionModel ListSelectionModel} for the range
     * of permitted values
     * @return a <code>ListSelectionModel</code> value
     */
    public static ListSelectionModel calcListSelectionModel(List vals, List sel, int selType) {
        if(logger.isDebugEnabled()) {
            logger.debug("Calculating");
        }
    
        ListSelectionModel lm = new org.enhydra.barracuda.core.comp.DefaultListSelectionModel();
        lm.setSelectionMode(selType);
        if(vals == null || sel == null || sel.size() == 0) {
            logger.debug("No selection");
            return lm;
        }
        for(int i = 0;i < vals.size(); i++) {
            Object key;
            Object rawKey, item = vals.get(i);
            if (logger.isDebugEnabled() && item != null) {
                logger.debug("Itemtype:" + item.getClass());
            } // end of if ((logger.isDebugEnabled())
        
            if(item instanceof ItemMap) {
                key = ((ItemMap)item).getKey();
            } else {
                key = item;
            }
            if (logger.isDebugEnabled() && key != null) {
                logger.debug("ItemKey:" + key);
                logger.debug("ItemKeyType:" + key.getClass());
            } // end of if ((logger.isDebugEnabled())

            if(sel.contains(key) || sel.contains(key.toString())) {
                logger.debug("Added interval " + i + " to " + i);
                lm.addSelectionInterval(i, i);
            }
        }
        if(logger.isDebugEnabled()) {
            for(int i = 0;i < vals.size(); i++) {
                if(lm.isSelectedIndex(i))
                    logger.debug("Index " + i + " is selected");        
            }
        }
        return lm;
    }

    /**
     * Creates a new <code>SelectFormElement</code> instance.
     *
     * @param ikey a <code>java.lang.String</code> value
     * @param iname a <code>java.lang.String</code> value
     * @param itype a <code>FormType</code> value
     * @param idefaultVal a <code>java.lang.Object</code> value
     * @param ivalidator a <code>FormValidator</code> value
     * @param iallowMultiples a <code>boolean</code> value
     */
    public SelectFormElement(java.lang.String ikey,
                             java.lang.String iname,
                             FormType itype,
                             java.lang.Object idefaultVal,
                             FormValidator ivalidator,
                             boolean iallowMultiples) {
        super(ikey, iname, itype, idefaultVal, ivalidator, iallowMultiples);
    
    }

    /**
     * Creates a new <code>SelectFormElement</code> instance.
     *
     * @param lm a <code>ListModel</code> value
     * @param ikey a <code>java.lang.String</code> value
     * @param iname a <code>java.lang.String</code> value
     * @param itype a <code>FormType</code> value
     * @param idefaultVal a <code>java.lang.Object</code> value
     * @param ivalidator a <code>FormValidator</code> value
     * @param iallowMultiples a <code>boolean</code> value
     */
    public SelectFormElement(ListModel lm, java.lang.String ikey,
                             java.lang.String iname,
                             FormType itype,
                             java.lang.Object idefaultVal,
                             FormValidator ivalidator,
                             boolean iallowMultiples) {
        super(ikey, iname, itype, idefaultVal, ivalidator, iallowMultiples);
        this.lm = lm;
    }

    /**
     * Creates a new <code>SelectFormElement</code> instance.
     *
     * @param lm a <code>ListModel</code> value
     * @param ikey a <code>java.lang.String</code> value
     * @param itype a <code>FormType</code> value
     * @param idefaultVal a <code>java.lang.Object</code> value
     * @param ivalidator a <code>FormValidator</code> value
     * @param iallowMultiples a <code>boolean</code> value
     */
    public SelectFormElement(ListModel lm, 
                             java.lang.String ikey,
                             FormType itype,
                             java.lang.Object idefaultVal,
                             FormValidator ivalidator,
                             boolean iallowMultiples) {
        super(ikey, itype, idefaultVal, ivalidator, iallowMultiples);
        this.lm = lm;
    }

    /**
     * Creates a new <code>SelectFormElement</code> instance.
     *
     * @param lm a <code>ListModel</code> value
     * @param lsm a <code>ListSelectionModel</code> value
     * @param ikey a <code>java.lang.String</code> value
     * @param iname a <code>java.lang.String</code> value
     * @param itype a <code>FormType</code> value
     * @param idefaultVal a <code>java.lang.Object</code> value
     * @param ivalidator a <code>FormValidator</code> value
     * @param iallowMultiples a <code>boolean</code> value
     */
    public SelectFormElement(ListModel lm, ListSelectionModel lsm, java.lang.String ikey,
                             java.lang.String iname,
                             FormType itype,
                             java.lang.Object idefaultVal,
                             FormValidator ivalidator,
                             boolean iallowMultiples) {
        super(ikey, iname, itype, idefaultVal, ivalidator, iallowMultiples);
        this.lm = lm;
        this.lsm = lsm;
    }
    /**
     * Creates a new <code>SelectFormElement</code> instance.
     *
     * @param ikey a <code>java.lang.String</code> value
     * @param itype a <code>FormType</code> value
     * @param idefaultVal a <code>java.lang.Object</code> value
     * @param ivalidator a <code>FormValidator</code> value
     */
    public SelectFormElement(java.lang.String ikey,
                             FormType itype,
                             java.lang.Object idefaultVal,
                             FormValidator ivalidator) {
        super(ikey, itype, idefaultVal, ivalidator);
    }

    /**
     * Creates a new <code>SelectFormElement</code> instance.
     *
     * @param lm a <code>ListModel</code> value
     * @param lsm a <code>ListSelectionModel</code> value
     * @param ikey a <code>java.lang.String</code> value
     * @param itype a <code>FormType</code> value
     * @param idefaultVal a <code>java.lang.Object</code> value
     * @param ivalidator a <code>FormValidator</code> value
     */
    public SelectFormElement(ListModel lm, ListSelectionModel lsm, java.lang.String ikey,
                             FormType itype,
                             java.lang.Object idefaultVal,
                             FormValidator ivalidator) {
        super(ikey, itype, idefaultVal, ivalidator);
        this.lm = lm;
        this.lsm = lsm;
    }

    /**
     * Describe <code>render</code> method here.
     *
     * @param context a <code>ViewContext</code> value
     * @return an <code>Object</code> value
     */
    public Object render(ViewContext context) {
        BSelect bs = new BSelect(getListModel());
        bs.setName(getName());
        bs.setSelectionModel(getListSelectionModel());
        return bs;
    }
}
