/*
 * MapDecorator.java
 *
 * Created on 9. Dezember 2002, 16:50
 */

package org.enhydra.barracuda.contrib.sam.models;

import java.util.*;
import java.text.*;

import org.enhydra.barracuda.core.comp.*;
import org.enhydra.barracuda.core.event.*;

import org.apache.log4j.*;

/** MapDecorator is a Decorator class using the well known decorator pattern.<br>
 * It modifies the return values of the {@link #getItem(TemplateDirective) } method of the
 * decorated model using a Map.
 * @author Stefan Armbruster
 * @version $Id: MapDecorator.java,v 1.6 2003/05/17 21:56:20 jacobk Exp $
 */
public class MapDecorator extends TemplateModelDecorator {
    
    protected static Logger logger = Logger.getLogger(MapDecorator.class);
    protected Map mapping;

	 public MapDecorator(TemplateModel tm) {
		super(tm);
	 }
    
    /** Creates a new instance of MapDecorator
     * @param tm decorated model
     * @param mapping Table containing the mappings, the keys represent the the keys in the {@link
     * #getItem(TemplateDirective)} method
     */
    public MapDecorator(TemplateModel tm, Map mapping) {
        super(tm);
			setMapping(mapping);
    }

	 public void setMapping(Map mapping) {
        this.mapping = mapping;
        if (logger.isDebugEnabled()) {
            Iterator iter = mapping.entrySet().iterator();
            while (iter.hasNext()) {
                Map.Entry entry = (Map.Entry)iter.next();
                logger.debug("Mapping table: key: " + entry.getKey() + " value: " + entry.getValue());
            }
        }
		}
    
    /** getItem first checks, if the is a matching entry in {@link #mapping }. If so,
     * the value of the decorated model is modified depending on the type of the mapper
     * entry. The following modifications are possible: <br>
     * <B>Lookup map, implements {@link java.util.Map }</B>
     * If the Map contains the value of the underlying model, it is replaced by the
     * value the the lookup map.
     * <B>formatting to a specific numeric format, implements {@link java.text.NumberFormat }</B>
     * <B>formatting to a specific date format, implements {@link java.text.DateFormat }</B>
     * <B>formatting to a user defined class, must implement UserMapping</B>
     * <B>returning a {@link org.enhydra.barracuda.core.comp.BAction }, implements {@link
     * org.enhydra.barracuda.core.event.ControlEvent }</B>
     * If the value of the mapping table is a ControlEvent, a BAction with the
     * ControlEvent as href is returned.
     * @param key key in the directive
     * @return modified return value of underlying model
     */    
//csc_030703.1    public Object getItem(String key) {
    public Object getItem(TemplateDirective td) {   //csc_030703.1
        String key = td.getKeyName();               //csc_030703.1

//csc_030703.1		logger.debug("getItem called, key : "+key);
//csc_030703.1        Object value = _templateModel.getItem(key);
		logger.debug("getItem called, td : "+td);   //csc_030703.1
        Object value = _templateModel.getItem(td);  //csc_030703.1
		logger.debug("_templateModels's getItem called, value : "+value);
        Object mapper = mapping.get(key);
		if (logger.isDebugEnabled()) {
			if (mapper==null) {	
				logger.debug("mapper is null");
			} else {
				logger.debug("mapper's class is " + mapper.getClass().getName());
			}
		}
        
        if (mapper instanceof Map) {
//csc_030703.1            value=mapItemAsLookup(key, _templateModel.getItem(key), (Map)mapper );
            value=mapItemAsLookup(key, _templateModel.getItem(td), (Map)mapper );   //csc_030703.1
        } else if (mapper instanceof NumberFormat) {
//csc_030703.1            value=mapItemAsNumberFormat(key, _templateModel.getItem(key), (NumberFormat)mapper );
            value=mapItemAsNumberFormat(key, _templateModel.getItem(td), (NumberFormat)mapper );    //csc_030703.1
        } else if (mapper instanceof DateFormat) {
//csc_030703.1            value=mapItemAsDateFormat(key, _templateModel.getItem(key), (DateFormat)mapper );
            value=mapItemAsDateFormat(key, _templateModel.getItem(td), (DateFormat)mapper );    //csc_030703.1
        } else if (mapper instanceof ControlEvent) {
//csc_030703.1            value=mapItemAsEvent(key, _templateModel.getItem(key), (ControlEvent) mapper);
            value=mapItemAsEvent(key, _templateModel.getItem(td), (ControlEvent) mapper);   //csc_030703.1
        } else if (mapper instanceof Link) {
//csc_030703.1            value=mapItemAsLink(key, (Link) mapper);
            value=mapItemAsLink(td, (Link) mapper);    //csc_030703.1
        } else if (mapper instanceof UserMapping) {
				value = mapItemAsUserMapping(key, (UserMapping)mapper);
        } else {
            if (mapper == null) {
                logger.info("mapper is null");
            } else {
                logger.error("invalid mapper class: " + mapper.getClass().getName());
            }
        }
        return value;
    }

	private Object mapItemAsUserMapping(String key, UserMapping um) {
				logger.debug("processing UserMapping");
            Object value=um.get(key, _templateModel);
				logger.debug("got from UserMapping: " + value);
				return value;
	}

    
    private Object mapItemAsLookup(String key, Object value, Map lookup) {
        Object ret = lookup.get(value);
        logger.debug("mapItemAsLookup " + key + " value " + value.getClass().getName()+ " ret " + ret);
        logger.debug("lookup: "+ lookup);
        if (ret == null) {
            ret = lookup.get("default");
            if (ret == null) ret = value;
        }
        return ret;
    }
    
    private Object mapItemAsNumberFormat(String key, Object value, NumberFormat format) {
        double d;
        if (value instanceof String) {
            d = Double.parseDouble((String)value);
        } else if (value instanceof Double) {
            d = ((Double)value).doubleValue();
        } else {
            logger.warn("no number conversion possible: key : " + key + " value: " + value);
            return value;
        }
        return format.format(d);
    }
    
    private Object mapItemAsDateFormat(String key, Object value, DateFormat format)  {
        java.util.Date d;
        if (value instanceof String) {
            DateFormat dateFormatIn = new SimpleDateFormat();
            try {
                d = dateFormatIn.parse((String)value);
            } catch (ParseException ex) {
                logger.error(ex.getMessage(), ex);
                return value;
            }
        } else if (value instanceof java.util.Date) {
            d = (java.util.Date)value;
        } else {
            logger.warn("no number conversion possible: key : " + key + " value: " + value);
            return value;
        }
        return format.format(d);
    }
    
    private Object mapItemAsEvent(String key, Object value, ControlEvent event)  {
        BAction ret = new BLink((String)value, event);
        ret.setParam(key, (String)value);
        return ret;
    }

//csc_030703.1    private Object mapItemAsLink(String key, Link link)  {
    private Object mapItemAsLink(TemplateDirective td, Link link)  {  //csc_030703.1
			logger.debug("building link to " + link.getEvent().getClass().getName());
			String label = link.getLabel();	
			if (!link.isLabelStatic()) {
//csc_030703.1				label = (String)_templateModel.getItem(label);
				label = (String)_templateModel.getItem(new TemplateDirective(td.getCommand(), td.getModelName(), label, td.getKeyData()));  //csc_030703.1
			}

			BAction ret;
			if (label == null) {
				ret = new BAction(link.getEvent());
			} else {
        		ret = new BLink(label, link.getEvent());
			}

			// add link parameters
			Iterator iter = link.getIterator();
			while (iter.hasNext()) {
				Map.Entry entry = (Map.Entry) iter.next();
				String k = (String) entry.getKey();
				StringBoolCompound c = (StringBoolCompound) entry.getValue();
				
				String v = c.getString();
				if (!c.getBool()) {
//csc_030703.1					v = (String) _templateModel.getItem( v);
					v = (String) _templateModel.getItem(new TemplateDirective(td.getCommand(), td.getModelName(), v, td.getKeyData()));    //csc_030703.1
				}
        		ret.setParam(k, v);
				logger.debug("setting link parameter " +k+ " to " + v);
				
			}
        return ret;
    }

	public static class Link {
		protected ControlEvent event;
		protected String label;
		protected boolean isLabelStatic;

		protected Map parameters = new HashMap();

		public Link(ControlEvent event, String label, boolean isLabelStatic) {
			this.event = event;
			this.label = label;
			this.isLabelStatic = isLabelStatic;
		}

		public String getLabel() {
			return label;
		}

		public ControlEvent getEvent() {
			return event;
		}

		public boolean isLabelStatic() {
			return isLabelStatic;
		}

		public Link addParameter(String key, String value, boolean isStatic) {
			parameters.put(key, new StringBoolCompound(value, isStatic));
			return this;
		}

		public Iterator getIterator() {
			return parameters.entrySet().iterator();
		}
	}

	public static class StringBoolCompound {
		protected String string;
		protected boolean bool;
		public  StringBoolCompound(String string, boolean bool) {
			this.string = string;
			this.bool = bool;
		}

		public String getString() {
			return string;
		}

		public boolean getBool() {
			return bool;
		}
	}

	/** Interface for handling user defined mappings */
	public static interface UserMapping {
		public Object get(String s, TemplateModel tm );
	}

}
