package org.enhydra.barracuda.contrib.sam.xmlform;

import java.util.*;
import java.lang.reflect.*;

import org.enhydra.barracuda.contrib.dbroggisch.repopulation.*;
import org.enhydra.barracuda.core.event.*;
import org.enhydra.barracuda.core.comp.*;
import org.enhydra.barracuda.core.forms.*;
import org.enhydra.barracuda.core.forms.validators.*;

import org.enhydra.barracuda.contrib.sam.xmlform.dtd.*;
import org.enhydra.barracuda.contrib.sam.data.*;

import org.apache.log4j.*;

/** Provide a FormMap based on a XML description */
public class XmlFormMap extends RepopulationFormMap {

	protected static Logger logger = Logger.getLogger(XmlFormMap.class.getName());

	public static final String GROUP = "GROUP";
	protected String formName;		
	protected Form form;

   /** instatiate a new XML Form. First the XML Form description is loaded by calling
    * {@link XmlFormFactory }. For each field, if required a validator is created and
    * the pre-set values are set according the DataObject
    * @param formName filename of the form, used to call {@link XmlFormFactory}
    * @param dataObject {@link DataObject} associated with the form
    * @param name internal name of the form
    */   
	public XmlFormMap(ControlEventContext context, String formName, String name, DataObject dataObject) 
	{
		this.formName = formName;
		setName(name);
		form = XmlFormFactory.get(context, this.formName);
		Iterator iter = form.getElementList().iterator();
		while (iter.hasNext()) {
			Element ele = (Element) iter.next();
			Object deflt = null;
			FormElement fe = getFormElement(ele, ele.getName());
			if (dataObject != null) {
				try {
					deflt = dataObject.get(ele.getName());
					if (deflt instanceof List) {

						// if DataObject's get method returns a List (instead of String), we know that this is a multi-value field
						List defltList = (List)deflt;
						SelectFormElement sfe = (SelectFormElement)fe;
						ListSelectionModel defaults = new DefaultListSelectionModel();
						defaults.setSelectionMode(ListSelectionModel.MULTIPLE_INTERVAL_SELECTION );

						// the list returned from get is a list of oids. These must be converted in select box indexes
						ListModel lm = sfe.getListModel();
						for (int i = 0; i< lm.getSize(); i++) {
							DefaultItemMap item = (DefaultItemMap)lm.getItemAt(i);
							if (defltList.contains(item.getKey())) defaults.addSelectionInterval(i,i);
						}
						sfe.setListSelectionModel(defaults);	
	
					} else {
						fe.setOrigVal(deflt);
						if (logger.isDebugEnabled()) logger.debug("default value for element " + ele.getName() + " is " + deflt);
					}
				} catch (DataObjectException ex) {
					logger.error(ex.getMessage(), ex);
				}
			}
			if (logger.isDebugEnabled()) logger.debug("Adding " + ele.getName() + " to formmap.");
			//fe.setOrigVal(xmlMap.get(ele.getName()));

			fe.setValidator( buildValidatorObject(ele.getValidator()));
			defineElement(fe);
      }
	}

	private static Map primitiveClasses = new Hashtable();
	
	static {
		primitiveClasses.put("int", Integer.TYPE);
		primitiveClasses.put("float", Float.TYPE);
		primitiveClasses.put("double", Double.TYPE);
		primitiveClasses.put("long", Long.TYPE);
		primitiveClasses.put("short", Short.TYPE);
		primitiveClasses.put("boolean", Boolean.TYPE);
		primitiveClasses.put("byte", Byte.TYPE);
		primitiveClasses.put("char", Character.TYPE);
	}

	private Class getClassWithPrimitive(String name)
		throws ClassNotFoundException
	{
		Class c = (Class) primitiveClasses.get(name);
		if (c==null) c = Class.forName(name);
		return c;
	}
	
	private Object initialize(Class c, String value) 
		throws NoSuchMethodException, InstantiationException, IllegalAccessException, InvocationTargetException
	{
		Class[] types = new Class[1];
		types[0] = String.class;
		Object[] para = new Object[1];
		para[0] = value;
		Constructor con = c.getConstructor(types);
		return con.newInstance(para);
	}
		

	/** recursivly called to construct a validator hierarchy. Internally, the reflection API is used here */
	protected FormValidator buildValidatorObject(Validator validator) 
	{
		if (validator == null) return null;

		List children = new Vector();

		//if (validator.getValidatorList() != null) {
			Iterator iter = validator.getValidatorList().iterator();
			while (iter.hasNext()) {
				Validator v = (Validator) iter.next();
				// get sub-validtors recursively
				children.add(buildValidatorObject(v));
			}
		//}
		try {
			// use reflection to instanciate validator
			logger.info("Trying to create Validator " + validator.getClassname());
			Class validatorClass = Class.forName(validator.getClassname());
			if (logger.isDebugEnabled()) logger.debug("Class successfully found");
		
			// find matching constructor, first the childvalidators and second the specific parameters are appendendA
			List para = validator.getParaList();
			
			int size = children.size() + para.size();
			if (logger.isDebugEnabled()) logger.debug("constructor will have " + size + " arguments");
			Class[] types = new Class[size];
			Object[] values = new Object[size];
			int counter = 0;
			// insert constructor parameter from children formvalidators
			iter = children.iterator(); 
			while (iter.hasNext()) {
				Object o = iter.next();
				types[counter] = FormValidator.class;
				values[counter++] = o;
			}
			// insert construcor parameter from <para> elements
			iter = para.iterator(); 
			while (iter.hasNext()) {
				Para p = (Para)iter.next();
				types[counter] = getClassWithPrimitive(p.getClassname());
				values[counter] = initialize(types[counter], p.getValue());
				counter++;
				if (logger.isDebugEnabled()) logger.debug("counter is " + counter);
			}

			Constructor c = validatorClass.getConstructor(types);
			if (logger.isDebugEnabled()) logger.debug("Constructor with " + children.size() + " arguments found");
			// invoke the constructor
			return (FormValidator)c.newInstance(values);
		} catch (Exception ex) {
			logger.error(ex.getMessage(), ex);
			return null;
		}
	}
		
   /** generate a FormElement for a given Element from the XML description */   
	protected FormElement getFormElement(Element ele, String name) 
	{
		FormElement fe = null;
		if (ele.getText() != null) 
			fe = new DefaultFormElement(name, FormType.STRING, null);
		else if ( ele.getTextarea() != null) {
			fe = new TextAreaFormElement(name, FormType.STRING, null);
			logger.debug("creating a textarea element"+name);
		}
		else if ( ele.getSelect() != null) {
			Select sel = ele.getSelect();
			DefaultListModel lm = new DefaultListModel();
			// call RMI to get select box contents
			if (sel.getXmlclass() != null) {
				logger.debug("class is " + sel.getXmlclass());
				try {
					Class c = Class.forName(sel.getXmlclass());
					Class[] params = new Class[0];
					Object[] args = new Object[0];
					Method method = c.getMethod(sel.getMethod(), params);
					Object result = method.invoke(null, args);
					logger.debug("result is " + result.getClass().getName());
					if (result instanceof List) {
						List l = (List)result;
						Iterator rmiIter = l.iterator();
						while (rmiIter.hasNext()) lm.add(rmiIter.next());
					}
				} catch (Exception e) {
					logger.error(e.getMessage(),e);
				}
			}
			Iterator subIter = ele.getSelect().getOptionList().iterator();
			while (subIter.hasNext()) {
				Option opt = (Option)subIter.next();
				lm.add(new DefaultItemMap(opt.getValue2(), opt.getValue()));
			}
			//fe = new SelectFormElement(lm, name, FormType.STRING, null, new NotNullValidator("Bitte whlen Sie im Feld " + ele.getLabel() + " eine Alternative aus."), false);
			boolean hasMulti = false;
			String m = sel.getMultivalue();
			if ( (m!=null) && (m.equals("true") || m.equals("yes"))) hasMulti = true;
			fe = new SelectFormElement(lm, name, FormType.STRING, null, null, hasMulti);
		} else if (ele.getHidden() != null) {
			fe = new DefaultFormElement(name, FormType.STRING, null);
		}
		return fe;
	}

	public Form getForm() 
	{
		return form;
	}

}
