/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site (http://www.enhydra.org/).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: AbstractBComponent.java,v 1.27 2003/04/14 17:04:56 christianc Exp $
 */
package org.enhydra.barracuda.core.comp;

import java.io.*;
import java.util.*;
import org.w3c.dom.*;
import org.w3c.dom.html.*;

import org.apache.log4j.*;

import org.enhydra.barracuda.core.util.dom.*;
import org.enhydra.barracuda.core.view.*;
import org.enhydra.barracuda.core.comp.renderer.*;
import org.enhydra.barracuda.plankton.*;
import org.enhydra.barracuda.plankton.data.*;

/**
 * This class provides the abstract implementation for BComponent.
 * Typically, however, you would extend from BComponent, not this class.
 */
public abstract class AbstractBComponent implements BContainer, StateMap {

    //public constants
    protected static final Logger logger = Logger.getLogger(AbstractBComponent.class.getName());

    //private constants    
    private static byte[] sep = System.getProperty("line.separator").getBytes();
    private static Map rfCompMap = new HashMap();

    //private vars
    protected BContainer parent = null;
    protected List children = new ArrayList();
    protected List stepChildren = null;
    protected boolean isStepChild = false;
    protected List views = new ArrayList();
    protected List tempViews = null;
    protected Object dvc = null;                        //default view context (opt)
    protected boolean preventCaching = false;
        
    protected boolean validated = false;
    protected StateMap statemap = new DefaultStateMap();


    //--------------- AbstractBComponent -------------------------
    public abstract void setName(String iname);
    public abstract String getName();
    public abstract void setVisible(boolean val);
    public abstract void setVisible(boolean val, boolean recurse);
    public abstract boolean isVisible();
    public abstract void setEnabled(boolean val);
    public abstract void setEnabled(boolean val, boolean recurse);
    public abstract boolean isEnabled();
    public abstract void setView(View view);
    public abstract void addView(View view);
    public abstract boolean removeView(View view);
    public abstract void removeAllViews();
    public abstract List getViews();
    public abstract void setAttr(Object attr, Object val);
    public abstract Object getAttr(Object attr);
    public abstract Map getAttrMap();
    public abstract void render(ViewContext vc) throws RenderException;
    public abstract boolean supports(ViewContext vc);

    /**
     * Set the default ViewContext. 
     *
     * @param dvc the default ViewContext
     */
    public void setDefaultViewContext(ViewContext idvc) {
        dvc = idvc;
    }

    /**
     * Get the default ViewContext. 
     *
     * @return the default ViewContext
     */
    public ViewContext getDefaultViewContext() {
        return (ViewContext) dvc;
    }
    
    /**
     * Tell the root component to prevent the DOM from caching 
     * when rendered to the client
     */
    public void setPreventCaching(boolean val) {
        getRootComponent().preventCaching = val;    
    }

    /**
     * Determine whether the root component is preventing DOM caching
     *
     * @return true if root component is preventing caching 
     */
    public boolean getPreventCaching() {
        return getRootComponent().preventCaching;    
    }
    
    public AbstractBComponent getRootComponent() {
        if (parent!=null && parent instanceof AbstractBComponent) {
            return ((AbstractBComponent) parent).getRootComponent();
        } else {
            return this;
        }
    }


    //--------------- Renderer -----------------------------------
    /**
     * This allows developers to install custom renderers for specific
     * classes of components. This should generally be done at startup.
     *
     * @param rf the RendererFactory to be registered
     * @param compCl the target document class 
     * @param domCl the target dom class 
     */
    public static void installRendererFactory(RendererFactory rf, Class compCl, Class domCl) {
        if (logger.isInfoEnabled()) logger.info("Installing renderer factory for comp:"+compCl+" for dom:"+domCl);
        synchronized (rfCompMap) {
            Map rfDomMap = (Map) rfCompMap.get(compCl);
            if (rfDomMap==null) {
                rfDomMap = new HashMap();
                rfCompMap.put(compCl, rfDomMap);
            }
            synchronized (rfDomMap) {
                rfDomMap.put(domCl, rf);
            }
        }
    }

    /**
     * This method allows a component to get a reference to the appropriate 
     * renderer. If an exact match does not exist, it will automatically 
     * look for renderers registered on parent classes.
     */
    public Renderer getRenderer(View view) throws NoSuitableRendererException {
        Node node = view.getNode();
        Class domCl = node.getClass();
        return getRenderer(domCl);    
    }
    
    public Renderer getRenderer(Class domCl) throws NoSuitableRendererException {
        Class cl = this.getClass();
        boolean registered = true;
        if (logger.isInfoEnabled()) logger.info("Looking up renderer factory for comp:"+cl+" for dom:"+domCl);
        
        //first try and find the dom map class
        Map rfDomMap = (Map) rfCompMap.get(cl);
        //..if it can't be located, look for the interfaces it implements
        if (rfDomMap==null) {
            List list = Classes.getAllInterfaces(domCl);
            Iterator it = list.iterator();
            while (it.hasNext()) {
                Class clint = (Class) it.next();
                if (logger.isDebugEnabled()) logger.debug("Looking for comp map interfaces:"+clint);
                rfDomMap = (Map) rfCompMap.get(clint);
                if (rfDomMap!=null) break;
            }
            registered = false;
        }
        //..if we still can't find it, look for parent classes
        if (rfDomMap==null) {
            if (logger.isDebugEnabled()) logger.debug("Looking for comp map superclass:"+cl.getSuperclass());
            rfDomMap = findComponentMap(cl.getSuperclass());
        }
        if (rfDomMap==null)    throw new NoSuitableRendererException("No renderer available for this component:"+cl);

        //now look for the actual factory
        RendererFactory rf = (RendererFactory) rfDomMap.get(domCl);
        //..if it can't be located, look for the interfaces it implements
        if (rf==null) {
            List list = Classes.getAllInterfaces(domCl);
            Iterator it = list.iterator();
            while (it.hasNext()) {
                Class clint = (Class) it.next();
                if (logger.isDebugEnabled()) logger.debug("Looking for dom map interfaces:"+clint);
                rf = (RendererFactory) rfDomMap.get(clint);
                if (rf!=null) break;
            }
            registered = false;
        }
        //..if we still can't find it, look for parent classes
        if (rf==null) {
            if (logger.isDebugEnabled()) logger.debug("Looking for dom map superclass:"+domCl.getSuperclass());
            rf = findRendererFactory(domCl.getSuperclass(), rfDomMap);
        }
        if (rf==null) throw new NoSuitableRendererException("No renderer available for this markup:"+domCl);

        //if the particular combination was not registered, register it now
        //so that the next time we look it up it'll be faster
        if (!registered) installRendererFactory(rf, cl, domCl);

        //return the renderer
        return rf.getInstance();
    }

    /**
     * Look for a factory that can render either this class or its parent
     * class.
     */
    protected Map findComponentMap(Class cl) {
        Map rfDomMap = (Map) rfCompMap.get(cl);
        if (rfDomMap==null && cl.getSuperclass()!=null) {
            rfDomMap = findComponentMap(cl.getSuperclass());
        }
        return rfDomMap;        
    }
    
    /**
     * Look for a factory that can render either this class or its parent
     * class.
     */
    protected RendererFactory findRendererFactory(Class cl, Map domMap) {
        RendererFactory rf = (RendererFactory) domMap.get(cl);
        if (rf==null && cl.getSuperclass()!=null) {
            rf = findRendererFactory(cl.getSuperclass(), domMap);
        }
        return rf;        
    }

    //--------------- Lifecycle ----------------------------------
    /**
     * Initialize cycle. The component should use this 
     * to perform any initialization. Note that the component
     * should be added to the overall component hierarchy before 
     * you invoke this method.
     */
    public void initCycle() {
        //initialize self
        
        //initialize all children
        if (children==null) children = new ArrayList();
        Iterator it = children.iterator();
        while (it.hasNext()) {
            Object child = it.next();
            if (child instanceof AbstractBComponent) ((AbstractBComponent) child).initCycle();
        }
    }
    
    /**
     * Destroy cycle. The component should use this cycle to
     * perform any special cleanup.
     */
    public void destroyCycle() {
        //cleanup all children
        if (children!=null) {   //csc_031202.1 - added condition: only do this if children!=null
            Iterator it = children.iterator();
            while (it.hasNext()) {
                Object child = it.next();
                if (child instanceof AbstractBComponent) ((AbstractBComponent) child).destroyCycle();
            }
        }
                    
        //cleanup self
        parent = null;
        children = null;    
        views = null;
        statemap = null;    //csc_041202.1 - seems like this should be getting cleared too...
    }




    //--------------- BContainer ---------------------------------
    /**
     * Set the parent container. Null indicates its the root.
     *
     * @param parent the parent container
     */
    public void setParent(BContainer iparent) {
        parent = iparent;
    }
    
    /**
     * Get the parent container. Returns null if it's the root.
     *
     * @return the parent container
     */
    public BContainer getParent() {
        return parent;    
    }
    
    /**
     * Add a child container to this one
     *
     * @param child the child container to be added
     */
    public void addChild(BContainer child) {
        //jrk_20021018.1 - added null check for children
        if ((child==null) ||
            (children!=null && children.contains(child))) return;
        //jrk_20021018.2 - unnecessary code since this case was already checked.
        /*if (children.contains(child)) {
            if (logger.isDebugEnabled()) logger.debug(this.toRef()+": skipping addChild since child is already added");
            return;            
        }*/
        if (logger.isDebugEnabled()) logger.debug(this.toRef()+": adding "+child);
        //jrk_20021018.3 - added null check for children
        if (children==null) children = new ArrayList();
        children.add(child);
        child.setParent(this);
        invalidate();
    }
    
    /**
     * Get a child container at a given index
     *
     * @param the index of the child container
     */
    public BContainer getChild(int index) {
        if (logger.isDebugEnabled()) logger.debug(this.toRef()+": getting idx:"+index);
        return (BContainer) children.get(index);
    }
    
    /**
     * Remove a child container from this one
     *
     * @param child the a child container to be removed
     * @return the child container that was removed
     */
    public BContainer removeChild(BContainer child) {
        if (child==null) return null;
        if (logger.isDebugEnabled()) logger.debug(this.toRef()+": removing "+child);
        return removeChild(children.indexOf(child));
    }

    /**
     * Remove a child container for a given index
     *
     * @param index of the child container to be removed
     * @return the child container that was removed
     */
    public BContainer removeChild(int index) {
        if (index<0) return null;
        if (logger.isDebugEnabled()) logger.debug(this.toRef()+": removing idx:"+index);
        BContainer child = (BContainer) children.remove(index);
        child.setParent(null);
        invalidate();
        return child;
    }

    /**
     * Remove all child containers from this one
     */
    public void removeAll() {
        if (logger.isDebugEnabled()) logger.debug(this.toRef()+": removing all children");
        if (children!=null) children.clear();
        invalidate();
    }
    
    /**
     * Determine whether this container has any children
     *
     * @return true if the container has child containers
     */    
    public boolean hasChildren() {
        return (children!=null && children.size()>0);
    }

    /**
     * Gets a list of all child BContainers. Note, this method 
     * returns a copy of the underlying child list
     *
     * @return a list of all child components.
     */    
    public List getChildren() {
        return new ArrayList(children);
    }

    /**
     * Determine whether this component is a step child
     *
     * @return true if this component is a step child
     */
    public boolean isStepChild() {
        return isStepChild;
    }
    
    /**
     * Add a step child; step children are automatically removed after
     * each render. 
     *
     * @param child the step child container to be added
     */
    public void addStepChild(BContainer child) {
        addStepChild(child, false);
    }
    
    /**
     * Add a step child; step children are automatically removed after
     * each render. 
     *
     * @param child the step child container to be added
     * @param inheritParentAttrs true if we want the step child to
     *      inherit the parent's settings for visibility/enabled
     */
    public void addStepChild(BContainer child, boolean inheritParentAttrs) {
        //jrk_20021018.4 - added null check for children
        if ((child==null) || 
            (stepChildren!=null && stepChildren.contains(child)) || 
            (children!=null && children.contains(child))) return;
        if (logger.isDebugEnabled()) logger.debug(this.toRef()+": adding stepchild "+child);
        this.addChild(child);
        if (stepChildren==null) stepChildren = new ArrayList();
        stepChildren.add(child);
        
        //csc_090501_start - the purpose of this fix is simple. When you add
        //a step child, its almost always because you're delegating behaviour
        //to a sub component. In these cases, that child's visibility/enabled
        //characteristics should generally match the parent (so if the parent
        //is disabled the child will be too).
        if (child instanceof BComponent) {
            BComponent chComp = (BComponent) child;
            chComp.isStepChild = true;
            if (inheritParentAttrs) {
                chComp.setVisible(this.isVisible());
                chComp.setEnabled(this.isEnabled());
            }
        }
        //csc_090501_finish

        //csc_041403.2 - added
        //if the child being added has no views, add in any views/temp views 
        //associated with the parent
        if (inheritParentAttrs && child instanceof BComponent) {
            BComponent bchild = (BComponent) child;
            if (!bchild.hasViews()) {
                Iterator it = getViews().iterator();
                while (it.hasNext()) {
                    View v = (View) it.next();
                    bchild.addView(v);
                    if (logger.isDebugEnabled()) logger.debug(this.toRef()+": adding view to stepchild "+child);
                }
                if (tempViews!=null) {
                    it = tempViews.iterator();
                    while (it.hasNext()) {
                        View v = (View) it.next();
                        bchild.addTempView(v);
                        if (logger.isDebugEnabled()) logger.debug(this.toRef()+": adding tempview to stepchild "+child);
                    }
                }
            }
        }
    }
    
    /**
     * Remove all step children
     */
    public void removeAllStepChildren() {
        if (logger.isDebugEnabled()) logger.debug(this.toRef()+": removing stepchildren");
        if (stepChildren!=null) {
            Iterator it = stepChildren.iterator();
            while (it.hasNext()) {
                BContainer tempChild = (BContainer) it.next();
                if (logger.isDebugEnabled()) logger.debug(this.toRef()+": removing stepchild "+tempChild);
                
                //csc_030802.1 - if we are removing a temp child make sure we destroy it
                //so that memory can get freed up. This bug reared its ugly head when I
                //was using a BTemplateViewHandler that was declared as an inner class, and which
                //used a model which itself was declared as an inner class of the view handler.
                //After the BTemplate renders, when it goes to remove its step children, it
                //must invoke destroyCycle on those or else they retain a reference to the 
                //BTemplate which prevents it from getting gc'd which in turn prevents the 
                //handler from getting gc'd which in turn runs you out of memory real fast on a
                //big result set. Nasty nasty nasty!
                if (tempChild instanceof BComponent) {
                    ((BComponent) tempChild).destroyCycle();
                }
                
//csc_102401.1 - oh was this a brutal typo...caused all kinds of problems when 
//you have a component hierarchy that was getting reused because the 
//temp child was never getting removed.
//                this.removeState(tempChild);
                this.removeChild(tempChild);
            }
        }
        stepChildren = null;
    }
    
    /**
     * Add a temporary view; temp views are automatically removed after
     * each render
     *
     * @param tview the temp view to be added
     */
    public void addTempView(View tview) {
        //jrk_20021018.5 - added null check for views
        if ((tview==null) || 
            (tempViews!=null && tempViews.contains(tview)) || 
            (views!=null && views.contains(tview))) return;
        if (logger.isDebugEnabled()) logger.debug(this.toRef()+": adding tempview "+tview);
        if (tempViews==null) tempViews = new ArrayList();
        tempViews.add(tview);
        
        //csc_041403.2 - added
        //in addition, when we add a temp view, add it to any step children 
        //that do not have a view. This makes it MUCH easier for a template model
        //to return nested components none of which may have a view yet...when
        //TemplateHelper assigns a temp view, it will stick to step children as well
        if (stepChildren!=null) {
            Iterator it = stepChildren.iterator();
            while (it.hasNext()) {
                BComponent tempChild = (BComponent) it.next();
                if (!tempChild.hasViews()) {
                    if (logger.isDebugEnabled()) logger.debug(this.toRef()+": adding tempview to stepchild");
                    tempChild.addTempView(tview);
                }
            }
        }
    }
    
    /**
     * Determine whether the component has any views (either regular or temp)
     *
     * @return true if the component has any views
     */    
    public boolean hasViews() {
        return ((views!=null && views.size()>0) ||
                (tempViews!=null && tempViews.size()>0));
    }

    /**
     * Invalidates the container and all parent containers above it.
     * This essentially indicates the components need to be laid out 
     * (in the server side world, this is equivalent to rendering)
     * again. After a component has been rendered, it will be marked 
     * valid again.
     */
    public void invalidate() {
        if (validated) {
            validated = false;
            if (getParent()!=null) getParent().invalidate();
        }
    }

    /**
     * Invalidates the container and all child containers below it.
     * Calling validate on the root of a tree effectively invalidates
     * the entire tree, forcing the whole tree to be revalidated the
     * next time it is rendered. This call is more expensive than an 
     * invalidate because it has to hit every node in the hierarchy.
     */
    public void validate() {
        validated = false;
        Iterator it = children.iterator();
        while (it.hasNext()) {
            BContainer child = (BContainer) it.next();
            child.validate();
        }
    }



    //-------------------- StateMap ------------------------------
    /**
     * set a property in this StateMap
     *
     * @param key the state key object
     * @param val the state value object
     */    
    public void putState(Object key, Object val) {
        statemap.putState(key,val);
    }
    
    /**
     * get a property in this StateMap
     *
     * @param key the state key object
     * @return the value for the given key
     */    
    public Object getState(Object key) {
        return statemap.getState(key);
    }
    
    /**
     * remove a property in this StateMap
     *
     * @param key the key object
     * @return the object which was removed
     */    
    public Object removeState(Object key) {
        return statemap.removeState(key);
    }
    
    /**
     * get a list of the keys for this StateMap
     *
     * @return a list the keys for this StateMap
     */    
    public List getStateKeys() {
        return statemap.getStateKeys();
    }
    
    /**
     * get a copy of the underlying Map
     *
     * @return a copy of the underlying state Map
     */    
    public Map getStateValues() {
        return statemap.getStateValues();    
    }



    //--------------- Utility Methods ----------------------------
    /**
     * Get the component reference
     *
     * @return a String representation of the component reference
     */
    public String toRef() {
//don't need the fully qualified class name since logger provides this    
//        return this.getClass().getName()+"@"+Integer.toHexString(this.hashCode());
        return "[@"+Integer.toHexString(this.hashCode())+"]";
    }
    
    /**
     * Get a String representation of the component
     *
     * @return a String representation of the component
     */
    public String toString() {
        return toString(getDefaultViewContext());
    }
    
    /**
     * Provide a String representation of the component. This will attempt to 
     * render the link as markup, which allows you to inline components along 
     * with text (a very useful feature!).
     *
     * @return a String representation of the component
     */
    public String toString(ViewContext vc) {
        if (vc==null) {
            String sname = getName();
            if (sname==null) sname = this.getClass().getName();
            int lpos = sname.lastIndexOf(".");
            if (lpos>-1) sname = sname.substring(lpos);        
            return sname+" (@"+Integer.toHexString(this.hashCode())+")";
        }

        //get the format type and the doc from the view context
        FormatType ft = vc.getViewCapabilities().getFormatType();
        Document doc = vc.getElementFactory().getDocument();

        //get any views associated with this link and then remove 
        //all views from the component
        List linkViews = this.getViews();
        this.removeAllViews();

        try {
            //get the appropriate renderer by looking up the
            //DOM class associated with the given format type
            Renderer r = this.getRenderer(ft.getDOMClass());

//csc_110501.1_start -             
/*
            //ask the renderer to create the default Node
            Node n = r.createDefaultNode(doc, vc);
            
            //create a view on the resulting node and add it in
            this.addTempView(new DefaultView(n));
*/
            //ask the renderer to create the default Node (Note that technically
            //this is a dangerous cast, since it assumes that we are actually an 
            //instance of BComponent...in reality, its probably ok, since we are
            //really only using AbstractBComponent for code hiding...if anyone
            //extends Barracuda components, they should always do so by extending
            //from BComponent, NOT by extending from AbstractBComponent. Unfortunately,
            //there's no way to enforce this, but in practice it probably won't happen 
            //very often, if at all, and if it does, at least its well documented... ;-)
            Node n = r.createDefaultNode(doc, (BComponent) this, vc); //csc_110501.1
            
            //if the component still needs a default view create one for it...
            if (!this.hasViews()) this.addTempView(new DefaultView(n));
//csc_110501.1_end

            //invalidate the component to ensure redraw
            this.invalidate();
            
            //now render the component
            this.render(vc);
            
            //now render the node
            StringWriter sw = new StringWriter(200);
            DefaultDOMWriter ddw = new DefaultDOMWriter();
            ddw.write(n, sw);
            String linkStr = sw.toString();
            String endln = System.getProperty("line.separator");
            if (linkStr.endsWith(endln)) linkStr = linkStr.substring(0,linkStr.length()-endln.length());
            return linkStr;
                        
        } catch (Exception e) {
            return ("Err building component markup string: "+e);
        } finally {
            //add the views back in if need be (resetting component
            //to original state)
            if (linkViews!=null) {
                Iterator it = linkViews.iterator();
                while (it.hasNext()) {
                    this.addView((View) it.next());                
                    if (validated) this.invalidate();
                }
            }
        }
    }

    public void printStackTrace(int depth, Logger logger) {
        printStackTrace(depth, logger, null);
    }
    
    public void printStackTrace(int depth, OutputStream out) {
        printStackTrace(depth, null, out);
    }
    
    /**
     * For debugging purposes. Print the basic structure of the
     * gateway.
     */
    protected void printStackTrace(int depth, Logger logger, OutputStream out) {
        if (depth<0) depth = 0;
        if (depth>25) depth = 25;
        String spaces = "                                                                              ";
        String inset = spaces.substring(0,depth*3);
        
        boolean stepchild = (parent!=null && 
                             ((AbstractBComponent) parent).stepChildren!=null && 
                             ((AbstractBComponent) parent).stepChildren.contains(this));
        print(logger, out, inset+this.getClass().getName() + "@" + Integer.toHexString(this.hashCode()) + (stepchild ? " (stepchild)" : ""));            

        //children
        print(logger, out, inset+"   children: "+(children==null ? "null" : ""));
        int cntr = -1;
        if (children!=null) {
            Iterator it = children.iterator();
            while (it.hasNext()) {
                BContainer cont = (BContainer) it.next();
                if (cont instanceof BComponent) {  
                    print(logger, out, inset+"      ["+(++cntr)+"]:");
                    ((BComponent) cont).printStackTrace(depth+2, logger, out);            
                } else {
                    print(logger, out, inset+"      ["+(++cntr)+"]"+cont.getClass().getName()+" (details unknown)");
                }
            }                
        }
        print(logger, out, inset+"   /end children");

        print(logger, out, inset+"/end @" + Integer.toHexString(this.hashCode()));
    }
    
    private static void print(Logger logger, OutputStream out, String s) {
        if (logger!=null) {
            logger.debug(s);
        } else if (out!=null) {
            try {
                out.write(s.getBytes());
                out.write(sep);
            } catch (IOException ioe) {}
        }
    }


}