/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site (http://www.enhydra.org/).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: BLink.java,v 1.19 2003/01/21 05:42:38 jacobk Exp $
 */
package org.enhydra.barracuda.core.comp;

import java.io.*;
import java.util.*;

import org.apache.log4j.*;
import org.w3c.dom.*;
import org.w3c.dom.html.*;

import org.enhydra.barracuda.core.comp.renderer.*;
import org.enhydra.barracuda.core.comp.renderer.html.*;
import org.enhydra.barracuda.core.event.*;
import org.enhydra.barracuda.core.event.events.*;
import org.enhydra.barracuda.core.util.dom.*;
import org.enhydra.barracuda.core.view.*;
import org.enhydra.barracuda.plankton.*;


/**
 * BLink is used to manipulate any element in a DOM template that
 * is capable of generating a URL request. In the case of HTML, this 
 * would typically include &lt;a&gt;, &lt;input&gt;, and &lt;button&gt; 
 * elements.
 *
 * <p>In most cases you will not actually need to bind the component
 * to a view in order to use it--if you return it from a model, this
 * will be done for you automatically. If however, you intend to use
 * the component <em>standalone</em> (ie. manually attaching it to a 
 * specific node in the DOM) or <em>inline</em> (ie. in a toString()), 
 * then you MUST BIND IT TO A VIEW before rendering, or an error will 
 * be generated. 
 *
 * <p>Because BLink may often be used for inling, it includes constructors 
 * that conveniently allow you to specify the a ViewContext (ie. instead of
 * an actual View)
 */
public class BLink extends BAction {

    //public vars
    protected static final Logger logger = Logger.getLogger(BLink.class.getName());
    
    //private vars
    protected String text = null;
    protected String target = null;
    protected boolean allowMarkupInText = false;    //csc_092701.1

    //--------------- Constructors -------------------------------
    /**
     * Public noargs constructor
     */
    public BLink() {}
    
    /**
     * Public constructor which creates the component and sets the text, 
     * and target values. This link will fire the default action event (unless
     * you manually specify an action). 
     *
     * @param text the text string that backs this component
     */
    public BLink(String itext) {
        if (itext!=null) setText(itext);
    }

    /**
     * Public constructor which creates the component and sets the text 
     * and action values. 
     *
     * @param text the text string that backs this component
     * @param actionUrl the action url to be fired (opt--if null, the default
     *      action specified in the template will be fired)
     */
    public BLink(String itext, String iactionUrl) {
        this(itext, iactionUrl, null);
    }

    /**
     * Public constructor which creates the component and sets the text 
     * and action values. This constructor takes a ViewContext object which 
     * will be used to create a default view for the component (ie. use this
     * constructor for inling a BLink)
     *
     * @param text the text string that backs this component
     * @param actionUrl the action url to be fired (opt--if null, the default
     *      action specified in the template will be fired)
     * @param dvc the default ViewContext (opt--its presence allows the 
     *         component to be rendered as markup in toString())
     */
    public BLink(String itext, String iactionUrl, ViewContext idvc) {
        if (idvc!=null) setDefaultViewContext(idvc);
        if (itext!=null) setText(itext);
        if (iactionUrl!=null) setAction(iactionUrl);
    }

    /**
     * Public constructor which creates the component and sets the text 
     * and action values. 
     *
     * @param text the text string that backs this component
     * @param actionEvent the action event to be fired (opt--if null, the default
     *        ActionEvent will be fired)
     */
    public BLink(String itext, ControlEvent iactionEvent) {
        this(itext, iactionEvent, null);
    }
    
    /**
     * Public constructor which creates the component and sets the text 
     * and action values. This constructor takes a ViewContext object which 
     * will be used to create a default view for the component (ie. use this
     * constructor for inling a BLink)
     *
     * @param text the text string that backs this component
     * @param actionEvent the action event to be fired (opt--if null, the default
     *        ActionEvent will be fired)
     * @param dvc the default ViewContext (opt--its presence allows the 
     *         component to be rendered as markup in toString())
     */
    public BLink(String itext, ControlEvent iactionEvent, ViewContext idvc) {
        if (idvc!=null) setDefaultViewContext(idvc);
        if (itext!=null) setText(itext);
        if (iactionEvent!=null) setAction(iactionEvent);
    }



    //--------------- Renderer -----------------------------------
    /**
     * Default component renderer factory registrations
     */
    static {
        HTMLRendererFactory rfHTML = new HTMLRendererFactory();
        installRendererFactory(rfHTML, BLink.class, HTMLElement.class);
    }

    /**
     * HTML RendererFactory
     */
    static class HTMLRendererFactory implements RendererFactory {
        public Renderer getInstance() {return new HTMLLinkRenderer();}
    }



    //--------------- BLink --------------------------------------
    /**
     * Set the text for this particular component
     *
     * @param text the text representation of this component
     */
    public void setText(String itext) {
        text = itext;
        invalidate();
    }
    
    /**
     * Get the text for this particular component
     *
     * @return the text for this particular component
     */
    public String getText() {
        return text;    
    }
    
    /**
     * Set the target for this particular component
     *
     * @param target the ext representation of the target
     */
    public void setTarget(String itarget) {
        target = itarget;
        invalidate();
    }
    
    /**
     * Get the target for this particular component
     *
     * @return the target for this particular component
     */
    public String getTarget() {
        return target;    
    }

    /**
     * Do we wish to allow markup in this text (defaults to false)
     *
     * @param val true if we wish to allow markup in the text
     */
    public void setAllowMarkupInText(boolean val) {
        allowMarkupInText = val;
        invalidate();
    }
    
    /**
     * See if we allow markup in the text
     *
     * @return true if we wish to allow markup in the text
     */
    public boolean allowMarkupInText() {
        return allowMarkupInText;    
    }
    
    
}