/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site (http://www.enhydra.org/).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: BList.java,v 1.17 2003/01/21 05:42:38 jacobk Exp $
 */
package org.enhydra.barracuda.core.comp;

import java.util.*;

import org.apache.log4j.*;
import org.w3c.dom.*;
import org.w3c.dom.html.*;

import org.enhydra.barracuda.core.comp.model.*;
import org.enhydra.barracuda.core.comp.renderer.*;
import org.enhydra.barracuda.core.comp.renderer.html.*;
import org.enhydra.barracuda.core.comp.renderer.xml.*;
import org.enhydra.barracuda.core.util.dom.DOMUtil;
import org.enhydra.barracuda.core.view.*;
import org.enhydra.barracuda.plankton.*;

/**
 * A BList component is used to render list data into a DOM 
 * template. It can be bound to a wide range of markup elements
 * (see HTMLListRenderer for details). 
 *
 * <p>In the case of BList, you will ALMOST ALWAYS need to manually
 * bind it to a View, unless you happen to be returning it from a model
 * (in which case this will be done for you automatically by cloning
 * the node which contained the directive)
 */
public class BList extends BComponent {

    //public vars
    protected static final Logger logger = Logger.getLogger(BList.class.getName());
    
    //private vars
    protected ListModel model = null;
    private LocalModelListener callback = null;

    //--------------- Constructors -------------------------------
    /**
     * Public noargs constructor
     */
    public BList() {}
    
    /**
     * Public constructor which creates the component and
     * binds it to a specific view
     *
     * @param model the specific model to back this component
     */
    public BList(ListModel imodel) {
        this(imodel, null); 
    }
    
    /**
     * Public constructor which creates the component and
     * binds it to a specific model. The component is also
     * bound to the specified view.
     *
     * <p>Null values may be passed in for any parameters, 
     * but if you do so you will need manually provide these
     * values (via the accessor methods) prior to actually 
     * rendering the component
     *
     * @param model the specific model to back this component
     * @param view the View the component should be bound to
     */
    BList(ListModel imodel, View iview) {
        if (imodel!=null) setModel(imodel);
        if (iview!=null) this.addView(iview);    
    }
    



    //--------------- Renderer -----------------------------------
    /**
     * Default component renderer factory registrations
     */
    static {
        HTMLRendererFactory rfHTML = new HTMLRendererFactory();
        installRendererFactory(rfHTML, BList.class, HTMLElement.class);
        installRendererFactory(rfHTML, BList.class, HTMLDocument.class);
    }

    /**
     * HTML RendererFactory
     */
    static class HTMLRendererFactory implements RendererFactory {
        public Renderer getInstance() {return new HTMLListRenderer();}
    }



    //--------------- BList -------------------------------
    /**
     * Set the model that backs the list. This causes the list to register
     * as a listener on the model, so any changes to it will be reflected in 
     * the list.
     *
     * @param model the model that backs the list
     */
    public void setModel(ListModel imodel) {

        //deregister if possible
        if (model!=null && callback!=null) {
            model.removeModelListener(callback);
        }
    
        //set the model    
        model = imodel;
        invalidate();
        
        //reregister if possible
        if (model!=null) {
            if (callback==null) callback = new LocalModelListener();
            model.addModelListener(callback);
        }
        
        //Q: how does this affect garbage collection? In other words, how do we
        //ensure that when we're done with this object it will get gc'd since the
        //model will still have a reference to it? This seems especially problematic
        //if we had multiple components sharing a model and one of them goes away.
        //
        //A: (I think) In the components destroyCycle, make sure the list model
        //gets set to null
    }
    
    /**
     * Get the model that backs the list
     *
     * @return the model that backs the list
     */
    public ListModel getModel() {
        if (model==null) setModel(new DefaultListModel());
        return model;    
    }
    
    /**
     * A convenience method that constructs a ListModel from an array of Objects 
     * and then applies setModel to it.
     *
     * @param list an array of Objects containing the items to display 
     *         in the list
     */
    public void setListData(Object[] list) {
        DefaultListModel lm = new DefaultListModel();
        if (list!=null) for (int i=0,max=list.length; i<max; i++) {
            lm.add(list[i]);
        }
        setModel(lm);    
    }
    
    /**
     * A convenience method that constructs a ListModel from an Iterator 
     * and then applies setModel to it.
     *
     * @param it an iterator of objects to display in the list
     */
    public void setListData(Iterator it) {
        DefaultListModel lm = new DefaultListModel();
        while (it.hasNext()) {
            lm.add(it.next());
        }
        setModel(lm);    
    }
    
    /**
     * Render a specific view for the component. 
     *
     * @param view View to be rendered
     * @param vc ViewContext for the client view
     * @throws RenderException if the particular View is not supported
     * @param list a List of all the views for this component
     */
/*
//021102.3_csc - removed, because now its in BComponent     
    protected void renderView (View view, ViewContext vc, int depth) throws RenderException {
        if (logger.isInfoEnabled()) logger.info("rendering comp:"+this.toRef()+" view:"+view);

        //actually render the view according to known interfaces
        try {
            Renderer r = getRenderer(view);
            r.renderComponent(this, view, vc);
            
        } catch (DOMException e) {
            logger.warn("DOM Error:", e);
            throw new DOMAccessException("Error rendering component in view:"+e, e);
        }
    }
*/



    //--------------- Lifecycle ----------------------------------
    /**
     * Destroy cycle. The component should use this cycle to
     * perform any special cleanup.
     */
    public void destroyCycle() {
        //default destroy
        super.destroyCycle();
    
        //we set the model to null so that the component can be 
        //garbage collected. If we don't do this, the model retains
        //a reference back to the component and so the component 
        //will never be freed up...
        setModel(null);        
    }




    //--------------- Utility ------------------------------------
/*    
    class LocalListDataListener implements ListDataListener {
        //for right now, just invalidate everything. In the future,
        //we could get a lot smarter and keep track of the ranges
        //changed and then only redraw those. This would greatly 
        //speed rendering, because we'd only need to redraw the parts
        //of the DOM that changed...
        public void intervalAdded(ListDataEvent e) {invalidate();}
        public void intervalRemoved(ListDataEvent e) {invalidate();}
        public void contentsChanged(ListDataEvent e) {invalidate();}
    }
*/    
    class LocalModelListener implements ModelListener {
        //get notified when one of the underlying models changes
        public void modelChanged(Model m) {
            invalidate();
        }
    }
}