/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site (http://www.enhydra.org/).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: TemplateDirective.java,v 1.4 2003/03/07 20:19:19 christianc Exp $
 */
package org.enhydra.barracuda.core.comp;

import java.util.*;
import java.io.*;
import java.net.*;

import org.w3c.dom.*;

/**
 * This class defines a TemplateDirective. A TemplateDirective is
 * created from a string (usually in the HTML class attribute) and
 * takes the form
 *     <strong><emph>Dir::&lt;command&gt;.&lt;model&gt;.&lt;key&gt;.&lt;props&gt;...</emph></strong>
 * where:
 * <ul>
 *    <li>&lt;command&gt; = the name of the command (custom directives are allowed)</li>
 *    <li>&lt;model&gt; = the name of the model</li>
 *    <li>&lt;key&gt; = the name of the key within the model</li>
 *    <li>&lt;data&gt; = any additional data (in any format) that may be associated with the directive</li>
 * </ul>
 * <p>As an example, the following are valid directives:
 * <ul>
 *    <li><strong><emph>Dir::Get_Data.UserData.FirstName</emph></strong></li>
 *    <li><strong><emph>Dir::Get_Data.UserData.LastName</emph></strong></li>
 *    <li><strong><emph>Dir::Get_Data.Iterate_Start.UserData..max=50&min=25&foo=bar</emph></strong></li>
 *    <li><strong><emph>Dir::Get_Data.Iterate_End</emph></strong></li>
 *    <li><strong><emph>Dir::Custom_Directive.UserData..do=something</emph></strong></li>
 * </ul>
 */
public class TemplateDirective {

    //supported directives commands
    public final static String ITERATE_START = "Iterate_Start";
    public final static String ITERATE_NEXT = "Iterate_Next";
    public final static String ITERATE_END = "Iterate_End";
    public final static String GET_DATA = "Get_Data";
    public final static String SET_ATTR = "Set_Attr";
    public final static String DISCARD = "Discard";

    protected String cmd = null;
    protected String modelName = null;
    protected String keyName = null;
    protected String keyData = null;

    protected TemplateDirective() {}
    
    //csc_030703.1 - added to make it easier to programatically create directives,
    //witout having to parse a String
    public TemplateDirective(String icmd, String imodelName, String ikeyName, String ikeyData) {
        cmd = icmd;
        modelName = imodelName;
        keyName = ikeyName;
        keyData = ikeyData;    
    }

    public String getCommand() {
        return cmd;
    }
    
    public String getModelName() {
        return modelName;
    }
    
    public String getKeyName() {
        return keyName;
    }
    
    public String getKeyData() {
        return keyData;
    }

    public String toString() {
        StringBuffer sb = new StringBuffer(50);
        sb.append("Dir::"+cmd);
        if (modelName!=null) {
            sb.append("."+modelName);
            if (keyName!=null) {
                sb.append("."+keyName);
                if (keyData!=null) {
                    sb.append("."+keyData);
                }
            }
        }
        return sb.toString();
    }

    public boolean equals(Object o) {
        if (o==null) return false;
        if (o==this) return true;
        if (!(o instanceof TemplateDirective)) return false;
        TemplateDirective td = (TemplateDirective) o;
        if (cmd!=null && td.getCommand()!=null && !cmd.equals(td.getCommand())) return false;
        if (modelName!=null && td.getModelName()!=null && !modelName.equals(td.getModelName())) return false;
        if (keyName!=null && td.getKeyName()!=null && !keyName.equals(td.getKeyName())) return false;
        if (keyData!=null && td.getKeyData()!=null && !keyData.equals(td.getKeyData())) return false;
        return true;    
    }

    /**
     * Given a space delimited string of directive commands, convert
     * each segment into a TemplateDirective and return all the 
     * valid directives in a List. Invalid directives are silently ignored.
     *
     * @param sourceStr a space delimited string of directive commands
     * @return a List of TemplateDirectives
     */    
    public static synchronized List getAllInstances(String sourceStr) {
        List directives = new ArrayList();
        StringTokenizer st = new StringTokenizer(sourceStr," ");
        while (st.hasMoreTokens()) {        
            try {
                directives.add(getInstance_private(st.nextToken()));    
            } catch (InvalidDirectiveException e) {}
        }
        return directives;
    }

    /**
     * Given a string of the form  
     * <strong><emph>Dir::&lt;command&gt;.&lt;model&gt;.&lt;key&gt;.&lt;props&gt;...</emph></strong>
     * convert it into a TemplateDirective and return that object.
     *
     * @param sourceStr a space delimited string of directive commands
     * @return the corresponding TemplateDirectives
     * @throws InvalidDirectiveException if the String cannot be converted
     */    
//csc_030703.1 - I see no reason why this needs to be synchronized
//csc_030703.1    public static synchronized TemplateDirective getInstance(String sourceStr) throws InvalidDirectiveException {
    public static TemplateDirective getInstance(String sourceStr) throws InvalidDirectiveException {   //csc_030703.1
        return getInstance_private(sourceStr);
    }

    private static TemplateDirective getInstance_private(String sourceStr) throws InvalidDirectiveException {
        //make sure it starts with Dir::
        if (sourceStr==null || !(sourceStr.startsWith("Dir::"))) throw new InvalidDirectiveException("Invalid Directive:"+sourceStr+"...Directive must start with 'Dir::'");
    
        //create a new TemplateDirective
        TemplateDirective td = new TemplateDirective();
    
        //create a String tokenizer and parse on periods    
        int pos0 = 5;
        int pos1 = -1;
        for (int cntr=0; cntr<4; cntr++) {
            pos1 = sourceStr.indexOf(".",pos0);
            if (pos1<0) pos1 = sourceStr.length();
            if (cntr==0 && pos0>-1 && pos1>-1 && pos0<pos1) {
                td.cmd = sourceStr.substring(pos0,pos1);
            } else if (cntr==1 && pos0>-1 && pos1>-1 && pos0<pos1) {
                td.modelName = sourceStr.substring(pos0,pos1);
            } else if (cntr==2 && pos0>-1 && pos1>-1 && pos0<pos1) {  
                td.keyName = sourceStr.substring(pos0,pos1);
            } else if (cntr==3 && pos0>-1 && pos0<sourceStr.length()) {
                td.keyData = sourceStr.substring(pos0);
            }
            pos0 = pos1+1;
        }
        
        return td;
    }

    
    public static void main(String[] args) {
        try {
            System.out.println ("TD 1="+TemplateDirective.getInstance("Dir::Get_Data.UserData.FirstName"));        
            System.out.println ("TD 2="+TemplateDirective.getInstance("Dir::Get_Data.UserData.LastName"));        
            System.out.println ("TD 3="+TemplateDirective.getInstance("Dir::Get_Data.Iterate_Start.UserData..max=50&min=25&foo=bar"));        
            System.out.println ("TD 4="+TemplateDirective.getInstance("Dir::Get_Data.Iterate_End"));        
            System.out.println ("TD 5="+TemplateDirective.getInstance("Dir::Custom_Directive.UserData..do=something"));        
            System.out.println ("TD 6="+TemplateDirective.getInstance("Dir::Veto_If_Null.UserData..Notes"));        
            
        } catch (InvalidDirectiveException e) {
            System.out.println ("Failure:"+e);
        }
    
    }

}