/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site (http://www.enhydra.org/).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: DOMComponentRenderer.java,v 1.9 2003/03/24 17:28:17 jacobk Exp $
 */
package org.enhydra.barracuda.core.comp.renderer;

import java.util.*;
import org.w3c.dom.*;
import org.w3c.dom.html.*;

import org.apache.log4j.Logger;

import org.enhydra.barracuda.core.comp.*;
import org.enhydra.barracuda.core.util.dom.*;
import org.enhydra.barracuda.plankton.*;

/**
 * This interface defines the methods needed to implement a Renderer.
 */
public abstract class DOMComponentRenderer implements Renderer {

    /**
     * The purpose of this method is for a renderer to provide 
     * a default node (if none exists). This is essentially where you
     * would say "if I need to render myself, and I'm not actually bound
     * to anything, what should the default markup look like if I had
     * to create it." 
     *
     * In most cases, its perfectly fine to just throw an 
     * UnsupportedFormatException (indicating that there is no default). The 
     * only renderer that currently implements this method is the 
     * HTMLLinkRenderer.
     *
     * @param doc the master Document which can be used to create elements
     *        from scratch
     * @param comp the component that we're dealing with for the current request
     * @param vc the view context for the current request
     * @return a default node (created from scratch)
     * @throws UnsupportedFormatException if the renderer has no default node
     */
//csc_110501.1    public Node createDefaultNode(Document doc, ViewContext vc) throws UnsupportedFormatException;
    public Node createDefaultNode(Document doc, BComponent comp, ViewContext vc) throws UnsupportedFormatException {  //csc_110501.1
        throw new UnsupportedFormatException("Cannot create default node");
    }

    /**
     * The purpose of this method is to add a child to a parent. In many
     * cases, this method is used to ensure that the resulting markup is 
     * valid, by inserting the appropriate markup in between the child and
     * the parent. In many cases, the generic Component renderer will be the 
     * only renderer that actually implements this method.
     *
     * @param parent the parent Node
     * @param child the child Node
     * @return the resulting parent node
     * @throws InvalidNodeException if teh child cannot be added to the parent
     */
    public Node addChildToParent(Node parent, Node child) throws InvalidNodeException {
        //eliminate the obvious
        if (parent==null || child==null) throw new InvalidNodeException("Invalid node: cannot add child:"+child+" to parent:"+parent);

        //make any adjustments specific to the markup
        //(--n/a--)
                
        //now add the child in
        if (child!=null) parent.appendChild(child);
        
        //return the parent
        return parent;
    }

    /**
     * This method should actually render the data from the component
     * into the view, taking into consideration the specified ViewContext.
     * Generally, every renderer will implement this method.
     *
     * @param comp the component to be rendered
     * @param view the view the component should be rendered in
     * @param vc the view context
     * @throws RenderException if unable to render the component in the 
     *        specified view
     */    
    public void renderComponent(BComponent comp, View view, ViewContext vc) throws RenderException {
        //plain jane components don't have any rendering by default
    }


    /**
     * This is just a debugging method to make it easy to show
     * the interfaces of a view being rendered...
     */
    public void showNodeInterfaces(View view, Logger extLogger) {    
        //get the node
        Node node = view.getNode();
        if (extLogger.isDebugEnabled()) extLogger.debug("node [id="+DOMUtil.getID(node)+"] implements the following interfaces:");
        Iterator it = Classes.getAllInterfaces(node).iterator();
        while (it.hasNext()) {
            Object o = it.next();
            if (extLogger.isDebugEnabled()) extLogger.debug("   "+o.toString());
        }
    }    
}