/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site (http://www.enhydra.org/).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: DefaultEventBroker.java,v 1.11 2003/01/21 05:42:52 jacobk Exp $
 */
package org.enhydra.barracuda.core.event;

import java.io.*;
import java.util.*;

import org.apache.log4j.*;

import org.enhydra.barracuda.plankton.data.*;

/**
 * <p>An EventBroker is responsible for two basic tasks:
 *
 * <ol>
 *    <li>it serves as a registry for listeners to express interest in
 *           in events (or to make themselves generally available)</li>
 *    <li>it serves as the central point in a system for dispatching events
 *          to interested parties.<li>
 * </ol>
 *
 * <p>Typically, only the ApplicationGateway will actually dispatch events 
 * using the event broker.
 *
 * <p>There are two ways to register for Events. One is to specify the
 * type of Event you're interested in. The other is simply to register
 * a listener for general availability...anything addressed to that
 * listener will be delivered. Note that when we're registering listeners,
 * we're really registering listener factories. This allows the broker
 * not to worry about syncronization issues, assuming that the factory
 * will take care of that detail.
 *
 * <p>Note that when adding listeners, the broker calculates aliases for
 * both listener IDs and events. This allows you to reference listeners
 * and events by using just the class name instead of having to use the
 * fully qualified class name. You can also add/remove aliases manually.
 * Note however, that aliases are never automatically removed by the system,
 * even when you remove listeners. This is because there is no way of knowing
 * how _many_ listeners correspond with a particular alias. So...we just leave 
 * the aliases registered. In practice, this shouldn't really impact things
 * as most listeners will only be deregistered at the end of the servlet 
 * lifecycle.
 *
 * <p>When there is more than one id for a given alias, the broker treats
 * it as an ambiguous case and returns no match for that particular alias.
 *
 * <p>The event extension is used by whatever instantiates the broker (usually
 * an ApplicationGateway) to define the particular event extension this
 * broker is handling).
 */
public class DefaultEventBroker implements EventBroker {

    //public constants
    protected static final Logger logger = Logger.getLogger(DefaultEventBroker.class.getName());

    //local vars
    protected Map idMap = new HashMap();
    protected Map eventMap = new HashMap();
    protected Map idXref = new HashMap();
    protected Map eventXref = new HashMap();
    protected String extension = null;
    protected DispatcherFactory dispatcherFactory = null;

    /**
     * Public constructor
     *
     * @param idispatcherFactory a reference to the dispatcher factory 
     *        to be used in creating dispatchers
     * @param iextension the extension to be associated with all events
     *        delivered through the dispatcher
     */
    public DefaultEventBroker(DispatcherFactory idispatcherFactory, String iextension) {
        dispatcherFactory = idispatcherFactory;
        if (dispatcherFactory==null) dispatcherFactory = new DefaultDispatcherFactory();
        extension = iextension;
    }
    
    /**
     * Return the event extension handled by this event broker
     *
     * @return the event extension associated with this broker
     */
    public String getEventExtension() {
        return extension;    
    }

    /**
     * register a listener id, so that events addressed to a 
     * specific listener can be delivered
     *
     * @param factory the listener factory to be added
     */    
    public void addEventListener(ListenerFactory factory) {
        //eliminate the obvious
        if (factory==null) return;
        String lid = factory.getListenerID();

        //add it to the idMap
        if (logger.isInfoEnabled()) logger.info("Adding listener id:"+lid);
        idMap.put(lid, factory);
        
        //get the aliases and add it to the id xref
        List lAliases = getAliases(lid);
        addAliases(lid, lAliases, idXref);
    }
    
    /**
     * add an event listener for a particular class of an event.
     * If the class referenced is not an instance of BaseEvent,
     * a ClassNotEventException will be thrown. Note that this 
     * method also registers the listener in the idMap as well.
     *
     * @param factory the listener factory to be added
     * @param event the specific class of event for which the factory is listening
     * @throws InvalidClassException if the event class does not implement BaseEvent
     */    
    public void addEventListener(ListenerFactory factory, Class event) throws InvalidClassException {
        //eliminate the obvious
        if (factory==null || event==null) return;
        if (!((BaseEvent.class).isAssignableFrom(event))) throw new InvalidClassException ("Class "+event.getName()+" is not a BaseEvent");
        if (logger.isInfoEnabled()) logger.info("Adding listener:"+factory+" for class:"+event);

        //add it to the eventMap
        List eventList = (List) eventMap.get(event);
        if (eventList==null) {
            eventList = new ArrayList();
            eventMap.put(event, eventList);
        }
        eventList.add(factory);

        //get the event aliases and add it to the event xref
        String eid = event.getName();
        List eAliases = getAliases(eid);
        addAliases(eid, eAliases, eventXref);
        
        //make sure it's also in the idMap
        String lid = factory.getListenerID();
        if (idMap.get(lid)==null) idMap.put(lid, factory);
        
        //get the aliases and add it to the id xref
        List lAliases = getAliases(lid);
        addAliases(lid, lAliases, idXref);
    }

    /**
     * remove a listener from general availability
     *
     * @param factory the listener factory to be removed
     */
    public void removeEventListener(ListenerFactory factory) {
        //eliminate the obvious
        if (factory==null) return;
        String lid = factory.getListenerID();
    
        //remove it from the idMap
        if (logger.isInfoEnabled()) logger.info("Removing listener id:"+lid);
        idMap.remove(lid);
    }
    
    /**
     * remove an event listener for specific types of events
     * If the class referenced is not an instance of BaseEvent,
     * a InvalidClassException will be thrown. Note that this 
     * method effectively deregisters the listener from the 
     * idMap as well.
     *
     * @param factory the listener factory to be removed
     * @param event the specific class of event for which the factory is listening
     * @throws InvalidClassException if the event class does not implement BaseEvent
     */
    public void removeEventListener(ListenerFactory factory, Class event) throws InvalidClassException {
        //eliminate the obvious
        if (factory==null || event==null) return;
        if (!((BaseEvent.class).isAssignableFrom(event))) throw new InvalidClassException ("Class "+event.getName()+" is not a BaseEvent");
        if (logger.isInfoEnabled()) logger.info("Removing listener:"+factory+" for class:"+event);
    
        //remove it from the eventMap
        List eventList = (List) eventMap.get(event);
        if (eventList!=null) {
            while (eventList.contains(factory)) {
                eventList.remove(factory);
            }
            if (eventList.size()<1) eventMap.remove(event);
        }
        
        //make sure we also remove it from the idMap
        String lid = factory.getListenerID();
        idMap.remove(lid);
    }
    
    /**
     * remove all references to an event listener, both for id and
     * for any event classes it has registered an interest in.
     *
     * @param factory the listener factory to be removed
     */
    public void purgeEventListener(ListenerFactory factory) {
        //eliminate the obvious
        if (factory==null) return;
        if (logger.isInfoEnabled()) logger.info("Purging listener:"+factory);

        //remove it from the idMap
        String lid = factory.getListenerID();
        idMap.remove(lid);

        //remove it from all lists in the eventMap
        Iterator it = eventMap.keySet().iterator();
        while (it.hasNext()) {
            Object eventKey = it.next();
            List eventList = (List) eventMap.get(eventKey);
            while (eventList.contains(factory)) {
                eventList.remove(factory);
            }
            if (eventList.size()<1) eventMap.remove(eventKey);
        }
    }
    
    /**
     * Get a specific listener based on listener ID
     *
     * @param id the listener id we're looking for
     * @return the ListenerFactory that matches that id
     */
    public ListenerFactory getEventListener(Object id) {
        //eliminate the obvious
        if (id==null) return null;
        if (logger.isInfoEnabled()) logger.info("Getting listener by id:"+id);
        
        //return a reference if it exists
        return (ListenerFactory) idMap.get(id);        
    }
    
    /**
     * Get a List of listeners for a type of event. Returns a copy 
     * of the broker's internal list, so you can do what you want 
     * with it. If the class referenced is not an instance of BaseEvent,
     * a InvalidClassException will be thrown. 
     *
     * @param event the event class we are looking for
     * @return a List of listeners that are interested in this class of event
     * @throws InvalidClassException if the event class does not implement BaseEvent
     */
    public List getEventListeners(Class event) throws InvalidClassException {
        //eliminate the obvious
        if (event==null) return null;
        if (!((BaseEvent.class).isAssignableFrom(event))) throw new InvalidClassException ("Class "+event.getName()+" is not a BaseEvent");
        if (logger.isInfoEnabled()) logger.info("Getting listeners for Class:"+event);
        
        //get the list and return it
        return (List) eventMap.get(event);
    }
    
    /**
     * Given a partial event class name, return the fully qualified class
     * name if it's possible to determine. If it is unknown, throw and
     * InvalidClassException. This method is primarily used by the 
     * ApplicationGateway to support event aliasing.
     *
     * @param eventStr the event name alias
     * @return the fully qualified event class name
     * @throws InvalidClassException if the eventStr cannot be unambiguously 
     *      matched to a class name
     */
    public String matchEventClass(String eventStr) throws InvalidClassException {
        if (eventStr==null) throw new InvalidClassException();  //csc_121302.1
//csc_121302.1        Object result = eventXref.get(eventStr);
        Object result = eventXref.get(eventStr.toLowerCase());  //csc_121302.1
        if (result==null || result instanceof List) throw new InvalidClassException();
        else return (String) result;
    }
        
    /**
     * Given a partial id name, return the fully qualified listener
     * ID if it's possible to determine. If it is unknown, throw and
     * InvalidClassException. This method is primarily used by the 
     * ApplicationGateway to support id aliasing.
     *
     * @param idStr the id name alias
     * @return the fully qualified listener id name
     * @throws InvalidClassException if the idStr cannot be unambiguously 
     *      matched to a listener id name
     */
    public String matchListenerID(String idStr) throws InvalidClassException {
        if (idStr==null) throw new InvalidClassException();     //csc_121302.1
//csc_121302.1        Object result = idXref.get(idStr);
        Object result = idXref.get(idStr.toLowerCase());        //csc_121302.1
        if (result==null || result instanceof List) throw new InvalidClassException();
        else return (String) result;
    }

    /**
     * The purpose of this method is to take a fully qualified
     * class name and return a list of aliases for it. For instance,
     * foo.blah.event.Test would generate the following aliases:
     * <ul>
     *   <li>foo.blah.event.Test</li>
     *   <li>blah.event.Test</li>
     *   <li>event.Test</li>
     *   <li>Test</li>
     * </ul>
     */
    protected List getAliases(String className) {
        List list = new ArrayList();
        
        //start by appending the full class name
        list.add(className);
        
        //now break it into pieces
        int spos = 0;
        int epos = className.length();
        while ((spos=className.indexOf('.',spos))>-1) {
            String alias = className.substring(spos+1);
            list.add(alias);
            spos+=1;
        }
    
        return list;    
    }

    /**
     * Manually register aliases for a given event (the aliases
     * will be determined automatically based on the class name
     * and the event ID)
     *
     * @param event the specific class of event we'd like to alias
     * @throws InvalidClassException if the event class does not implement BaseEvent
     */
    public void addEventAlias(Class event) throws InvalidClassException {
        if (event==null) return;
        if (!((BaseEvent.class).isAssignableFrom(event))) throw new InvalidClassException ("Class "+event.getName()+" is not a BaseEvent");
        
        //start by aliasing off the class name
        addEventAlias(event, event.getName());    
        
        //if the event id is different than the class name alias off that too
        try {
            BaseEvent be = (BaseEvent) event.newInstance();
            if (!be.getEventID().equals(event.getName())) addEventAlias(event, be.getEventID());            
        } catch (Exception e) {}
    }

    /**
     * Manually add an alias for a given event. Note that
     * the alias parameter will be converted into all possible 
     * aliases based on '.' delimiters.
     *
     * @param event the specific class of event we'd like to alias
     * @param alias the alias for this event
     * @throws InvalidClassException if the event class does not implement BaseEvent
     */
    public void addEventAlias(Class event, String alias) throws InvalidClassException {
        if (event==null || alias==null) return;
        if (!((BaseEvent.class).isAssignableFrom(event))) throw new InvalidClassException ("Class "+event.getName()+" is not a BaseEvent");
        List lAliases = getAliases(alias);
        addAliases(event.getName(), lAliases, eventXref);
    }

    /**
     * Given an id and a list of aliases, add them to the 
     * specified xref
     */
    protected void addAliases(String id, List aliases, Map xref) {
        Iterator it = aliases.iterator();
        while (it.hasNext()) {
            //get the alias and see if it's already in the map
//csc_121302.1            Object alias = it.next();
            String alias = ((String) it.next()).toLowerCase();   //csc_121302.1
            Object ref = xref.get(alias);

            //if it's not, just add it
            if (ref==null) {
                xref.put(alias, id);
            
            //if it is...
            } else {
                //if the ref is not a list, create one and add it to the list
                List list = null;
                if (ref instanceof List) {
                    list = (List) ref;
                } else {                
                    //first make sure this isn't a dupe (if it is, just continue)
                    if (id.equals(ref)) continue;
            
                    //go ahead and create the list
                    list = new ArrayList();
                    list.add(ref);
                    xref.put(alias, list);
                }
                
                //now add the reference to the list if the id is not
                //already in there (this prevents dupes)
                if (!list.contains(id)) list.add(id);
            }
        }
    }

    public static void main (String[] args) {
        DefaultEventBroker eb = new DefaultEventBroker(null, ".event");
        eb.runTest();
    }
    
    public void runTest() {
        //test 1
        if (true) {
            String lid = "foo.blah.event.Test";
            List lAliases = getAliases(lid);
            addAliases(lid, lAliases, idXref);
        }
        //test 2
        if (true) {
            String lid = "foo.blah.event.Blarney";
            List lAliases = getAliases(lid);
            addAliases(lid, lAliases, idXref);
        }
        //test 3
        if (true) {
            String lid = "foo.blah.event2.Test";
            List lAliases = getAliases(lid);
            addAliases(lid, lAliases, idXref);
            addAliases(lid, lAliases, idXref);
            addAliases(lid, lAliases, idXref);
        }
        
        //now see what we have
        System.out.println ("");
        System.out.println ("Check @ point 1:");
        CollectionsUtil.printStackTrace(idXref, System.out);
        

        //now see if we get matches
        String id = null;
        try {
            id = "Blarney";                    //should work
            String result = matchListenerID(id);    
            System.out.println ("id:"+id+" result:"+result);    
        } catch (InvalidClassException e) {System.out.println ("Whoops..."+id+" didn't work!");}
        try {
            id = "event.Test";                //should work
            String result = matchListenerID(id);    
            System.out.println ("id:"+id+" result:"+result);    
        } catch (InvalidClassException e) {System.out.println ("Whoops..."+id+" didn't work!");}
        try {
            id = "Test";                    //should fail
            String result = matchListenerID(id);    
            System.out.println ("Whoops...didn't work!");    
        } catch (InvalidClassException e) {System.out.println ("Got error for "+id+" as expected");}
    }
    
    /**
     * <p>Dispatch a queue of events. Generally, the queue will only
     * contain one event, however, if you ever need to dispatch
     * multiple events at once, the broker can handle it. All the real 
     * dispatching work is carried out by the underlying event
     * dispatcher.
     *
     * <p>The event queue you pass in should contain several pieces of 
     * state information:
     *
     * <ul>
     *    <li>DefaultEventDispatcher.DEFAULT_RESPONSE_EVENT (BaseEvent)</li>
     *    <li>DefaultBaseEventListener.HTTP_SERVLET_REQUEST (HttpServletRequest) - if dispatching from a servlet</li>
     *    <li>DefaultBaseEventListener.HTTP_SERVLET_RESPONSE (HttpServletResponse) - if dispatching from a servlet</li>
     * </ul>
     *
     * @param context the EventContext to be dispatched
     * @throws EventException
     */
    public void dispatchEvent(EventContext context) throws EventException {
        //eliminate the obvious
        if (context==null) return;

        //dispatch the event
        if (logger.isInfoEnabled()) logger.info("Dispatching event:"+context);
        dispatcherFactory.getInstance().dispatchEvent(this, context);
        if (logger.isInfoEnabled()) logger.info("Dispatch complete!");
    }
}
