/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site (http://www.enhydra.org/).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: DefaultEventContext.java,v 1.18 2003/01/21 05:42:52 jacobk Exp $
 */
package org.enhydra.barracuda.core.event;

import java.io.*;
import java.util.*;
import javax.servlet.*;
import javax.servlet.http.*;

import org.apache.log4j.*;

import org.enhydra.barracuda.core.view.*;
import org.enhydra.barracuda.plankton.data.*;

/**
 * <p>The context contains information about the event (event, 
 * queue, plus request and response info if appropriate). The
 * context also implements StateMap, so you can use it to pass 
 * information between event handlers. The scope of the context 
 * is only for the duration of an event dispatch cycle (ie. from 
 * Http Request to Http Response)
 *
 * <p>Note that because DefaultEventContext implements ViewEventContext
 * it can safely be cast to its two parent interfaces: EventContext and
 * ControlEventContext. This is in fact what the DefaultEventDispatcher
 * actually does to control access to the appropriate information 
 * (ie. a ControlEvent handler should get a reference to an ControlEventContext,
 * NOT an ViewEventContext, since the control handlers shouldn't have
 * access to the HttpServletResponse object.)
 */
public class DefaultEventContext implements ViewEventContext {

    //public vars
    protected static final Logger logger = Logger.getLogger(DefaultEventContext.class.getName());

    //private vars
    private static final String CONTEXT_ID = "$ctxid";
    private static final String STATE = "_STATE";
    private static final String LOCAL_OR = "_LOCAL_OR";
    private StateMap statemap = null;
    private static long uid = System.currentTimeMillis();
    
    /**
     * Public constructor. 
     *
     * @param queue the event dispatch queue
     * @param config the servlet config structure
     * @param req the HttpServletRequest
     * @param resp the HttpServlet Response
     * @param defaultResponseEvent the default response event to be used 
     *        if none is generated by the control event handlers
     */    
    public DefaultEventContext(DispatchQueue queue, ViewCapabilities vc, ServletConfig config, HttpServletRequest req, HttpServletResponse resp, HttpResponseEvent defaultResponseEvent) {
        restoreContext(req);
        this.putState(EventContext.DISPATCH_QUEUE, queue);    
        this.putState(EventContext.VIEW_CAPABILITIES, vc);    
        this.putState(ControlEventContext.SERVLET_CONFIG, config);    
        this.putState(ControlEventContext.HTTP_SERVLET_REQUEST, req);    
        this.putState(ViewEventContext.HTTP_SERVLET_RESPONSE, resp);    
        this.putState(DefaultEventDispatcher.DEFAULT_RESPONSE_EVENT, defaultResponseEvent);
    }

    /**
     * Get the underlying BaseEvent
     *
     * @return the underlying BaseEvent
     */
    public BaseEvent getEvent() {
        return (BaseEvent) this.getState(EventContext.BASE_EVENT);    
    }

    /**
     * Get the underlying DispatchQueue
     *
     * @return the underlying DispatchQueue
     */
    public DispatchQueue getQueue() {
        return (DispatchQueue) this.getState(EventContext.DISPATCH_QUEUE);    
    }
    
    /**
     * Get the underlying ViewCapabilities
     */
    public ViewCapabilities getViewCapabilities() {
        return (ViewCapabilities) this.getState(EventContext.VIEW_CAPABILITIES);
    }
    
    /**
     * Get the associated ServletConfig structure
     *
     * @return the associated ServletConfig structure
     */
    public ServletConfig getConfig() {
        return (ServletConfig) this.getState(ControlEventContext.SERVLET_CONFIG);    
    }
    
    /**
     * Get the associated HttpServletRequest
     *
     * @return the associated HttpServletRequest
     */
    public HttpServletRequest getRequest() {
        return (HttpServletRequest) this.getState(ControlEventContext.HTTP_SERVLET_REQUEST);    
    }

    /**
     * Get the associated HttpServletResponse
     *
     * @return the associated HttpServletResponse
     */
    public HttpServletResponse getResponse() {
        return (HttpServletResponse) this.getState(ViewEventContext.HTTP_SERVLET_RESPONSE);    
    }
    
    //csc_082302.2 - rewritten to persist context using a unique id (per request)
    //in a Global Weak Repository, rather than persisting context in user session,
    //which inherantly had problems if multiple frames on a client were firing 
    //different events at the same time. The down side of this fix is that the
    //redirected url shows in the browser with something like this tacked onto 
    //the url: $ctxid=_1030135624563
    /**
     * The event context must be able to persist its state 
     * in the user's session. NOTE: this method is primarily 
     * intended for use by the EventDispatcher class when the
     * system needs to persist context information over multiple
     * request-response cycels (ie. to handle ClientSideRedirectExceptions).
     * As such, developers using the event model in applications
     * should never really need to call this method.
     *
     * @param re the client side redirect exception that is the reason
     *  for this call to persistContext
     */
    public void persistContext(ClientSideRedirectException re) {
        if (logger.isInfoEnabled()) logger.info("persisting context to prepare for client side redirect...");

/*
        SessionServices.getSession(getRequest()).setAttribute(STATE, statemap);
        //csc_022502.1_start
        ObjectRepository lr = ObjectRepository.getLocalRepository();
        List list = lr.getStateKeys();
        if (list!=null && list.size()>0) {
            StateMap sm = new DefaultStateMap();
            SessionServices.getSession(getRequest()).setAttribute(LOCAL_OR, sm);
            Iterator it = list.iterator();
            while (it.hasNext()) {
                Object key = it.next();
                Object val = lr.getState(key);
                sm.putState(key, val);
            }
        }
        //csc_022502.1_end
*/        
        //grab a reference to the weak global repository (we use a weak reference
        //here just in case the client hits stop or they do not support redirects;
        //since we aren't putting the information in the session, we have to make sure
        //it gets cleaned up, and this is the easiest way to do that)
        ObjectRepository wgor = ObjectRepository.getWeakGlobalRepository();
        String id = getUID();
        if (logger.isDebugEnabled()) logger.debug("...getting uid:"+id);
        
        //persist the statemap and local repository information in the weak global
        //repository (note that the NS is what will keep this unique)
        //...statemap
        wgor.putState(CONTEXT_ID+id+STATE, statemap);
        if (logger.isDebugEnabled()) logger.debug("...saved statemap, key = "+CONTEXT_ID+id+STATE);        
        //...local or
        ObjectRepository lr = ObjectRepository.getLocalRepository();
        List list = lr.getStateKeys();
        if (list!=null && list.size()>0) {
            StateMap sm = new DefaultStateMap();
            wgor.putState(CONTEXT_ID+id+LOCAL_OR, sm);
            if (logger.isDebugEnabled()) logger.debug("...saved lr, key = "+CONTEXT_ID+id+LOCAL_OR);        
            Iterator it = list.iterator();
            while (it.hasNext()) {
                Object key = it.next();
                Object val = lr.getState(key);
                sm.putState(key, val);
                if (logger.isDebugEnabled()) logger.debug("......saved lr items: key="+key+" val="+val);        
            }
        }
        
        //finally, adjust the redirect url to ensure that we can extract the NS id
        //from the url when it comes time to restore the context
        String url = re.getRedirectURL();
        int spos = url.indexOf("?");
        String sep = (spos>-1 ? "&" : "?");
        re.setRedirectURL(url+sep+CONTEXT_ID+"="+id);
        if (logger.isDebugEnabled()) logger.debug("...adjusting redirect URL from:"+url+" to:"+re.getRedirectURL());
    }
    
    //csc_082302.2 - rewritten along with above method
    /**
     * The event context must also be able to restore state
     * from the user's session. Calling this method effectively
     * removes the state from the session. 
     */
    public void restoreContext(HttpServletRequest req) {
        if (logger.isInfoEnabled()) logger.info("see if we need to restore context...");

        //start by trying to get the NS id from the request
        String id = req.getParameter(CONTEXT_ID);
        if (logger.isDebugEnabled()) logger.debug("...looking for context id in req:"+id);
        if (id!=null) {
            if (logger.isDebugEnabled()) logger.debug("...found persisted context! attempting to restore...");

            //see if we can find the id in the weak global or
            ObjectRepository wgor = ObjectRepository.getWeakGlobalRepository();
            //...statemap
            String stkey = CONTEXT_ID+id+STATE;
            Object o = wgor.getState(stkey);
            if (logger.isDebugEnabled()) logger.debug("......looking for statemap in wgor, key:"+stkey+" found:"+o);
            if (o!=null && o instanceof StateMap) statemap = (StateMap) o;
            else logger.warn("......expected statemap context missing!");
            wgor.removeState(stkey);
            //...local repository
            stkey = CONTEXT_ID+id+LOCAL_OR;
            o = wgor.getState(stkey);
            if (logger.isDebugEnabled()) logger.debug("......looking for lr in wgor, key:"+stkey+" found:"+o);
            if (o!=null && o instanceof StateMap) {
                StateMap sm = (StateMap) o;
                if (sm!=null) {
                    List list = sm.getStateKeys();
                    if (list!=null) {
                        ObjectRepository lr = ObjectRepository.getLocalRepository();
                        Iterator it = list.iterator();
                        while (it.hasNext()) {
                            Object key = it.next();
                            Object val = sm.getState(key);
                            lr.putState(key, val);
                            if (logger.isDebugEnabled()) logger.debug(".........retrieved lr items: key="+key+" val="+val);        
                        }
                    }
                }
            }
            wgor.removeState(stkey);
        } else {
            if (logger.isDebugEnabled()) logger.debug("...no persisted context");
        }

        if (statemap==null) statemap = new DefaultStateMap();
/*
        HttpSession session = SessionServices.getSession(req, false);
        if (session!=null) {
            statemap = (StateMap) session.getAttribute(STATE);    
            session.removeAttribute(STATE);
        }
        if (statemap==null) statemap = new DefaultStateMap();
        //csc_022502.1_start
        if (session!=null) {
            StateMap sm = (StateMap) session.getAttribute(LOCAL_OR);
            if (sm!=null) {
                List list = sm.getStateKeys();
                if (list!=null) {
                    ObjectRepository lr = ObjectRepository.getLocalRepository();
                    Iterator it = list.iterator();
                    while (it.hasNext()) {
                        Object key = it.next();
                        Object val = sm.getState(key);
                        lr.putState(key, val);
                    }
                }
                session.removeAttribute(LOCAL_OR);
            }
        }
        //csc_022502.1_end
*/
    }
    
    private synchronized static String getUID() {
        return "_"+(++uid);
    }

    
    
    //-------------------- StateMap ------------------------------
    /**
     * set a property in this StateMap
     *
     * @param key the state key object
     * @param val the state value object
     */    
    public void putState(Object key, Object val) {
        statemap.putState(key,val);
    }
    
    /**
     * get a property in this StateMap
     *
     * @param key the state key object
     * @return the value for the given key
     */    
    public Object getState(Object key) {
        return statemap.getState(key);
    }
    
    /**
     * remove a property in this StateMap
     *
     * @param key the key object
     * @return the object which was removed
     */    
    public Object removeState(Object key) {
        return statemap.removeState(key);
    }
    
    /**
     * get a list of the keys for this StateMap
     *
     * @return a list the keys for this StateMap
     */    
    public List getStateKeys() {
        return statemap.getStateKeys();
    }
    
    /**
     * get a copy of the underlying Map
     *
     * @return a copy of the underlying state Map
     */    
    public Map getStateValues() {
        return statemap.getStateValues();    
    }
    
}
