/*
 * Enhydra Java Application Server Project
 *
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site (http://www.enhydra.org/).
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific terms governing rights and limitations
 * under the License.
 *
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 * Christian Cryder, Diez B. Roggisch
 *
 * $Id: FormElement.java,v 1.14 2003/01/21 05:42:53 jacobk Exp $
 */
package org.enhydra.barracuda.core.forms;

import java.util.*;
import javax.servlet.*;
import javax.servlet.http.*;

import org.enhydra.barracuda.plankton.data.DefaultStateMap;
import org.enhydra.barracuda.plankton.data.StateMap;

/**
 * <p>This interfaces defines the methods a class needs to implement
 * to act as a FormElement.
 *
 * <p>A FormElement defines how an element in a FormMap should
 * be mapped to a first class java object. There are several
 * key pieces of information required:
 *
 * <ul>
 *         <li>key - used to retrieve the target value from the incoming
 *            data source (either a ServletRequest or a StateMap)</li>
 *         <li>type - the type of data object the incoming value should be
 *            mapped to, as defined by the FormType class (String, Boolean,
 *            Integer, Date, Long, Short, Double, Float)</li>
 *         <li>default val - the default val to be used if the key value is
 *            null in the incoming data source (may be null)</li>
 *         <li>validator - any validators associated with this particular
 *            form elmenent (may be null)</li>
 *         <li>allow multiples - the servlet interface allows multiple params
 *            with the same key. This value indicates whether or not that is
 *            acceptable for a key. If so, then when multiple values are
 *            encountered they element will actually contain a List of values
 *            of the proper type</li>
 * </ul>
 *
 * <p>Once we have this information, we basically have enough data to map
 * a String value in the data source to a first class Java object in the
 * element.
 */
public interface FormElement {

    /**
     * Set the key value for this form element
     *
     * @param key the key value for this form element
     */
    public void setKey(String key);

    /**
     * Get the key value for this form element
     *
     * @return the key for this form element
     */
    public String getKey();

    /**
     * Set the name of this form element
     *
     * @param name the name value of this form element
     */
    public void setName(String name);

    /**
     * Get the name value of form element
     *
     * @return the name of this form element
     */
    public String getName();

    /**
     * Set the FormType for this form element
     *
     * @param type the FormType for this form element
     */
    public void setType(FormType type);
    
    /**
     * Get the FormType for this for element
     *
     * @return the FormType for this form element
     */
    public FormType getType();

    /**
     * Set the default value for this form element
     *
     * @param defaultVal the FormType for this form element
     */
    public void setDefaultVal(Object defaultVal);
    
    /**
     * Get the default value for this form element
     *
     * @return the default value for this form element (may be null)
     */
    public Object getDefaultVal();

    /**
     * Set whether or not this element allows multiple values
     *
     * @param val true if the element allows multiples
     */
    public void setAllowMultiples(boolean val);

    /**
     * Does this element allow multiple values
     *
     * @return true if this element allows multiple values
     */
    public boolean allowMultiples();

    /**
     * Set the FormValidator for this form element
     *
     * @param validator the FormValidator for this form element
     */
    public void setValidator(FormValidator validator);

    /**
     * Get the default FormValidator for this form element
     *
     * @return the validator for this form element (may be null)
     */
    public FormValidator getValidator();

    /**
     * Set the original value for this element
     *
     * @param origVal the original value
     */
    public void setOrigVal(Object iorigVal);

    /**
     * Get the original value for this element
     *
     * @return the original value for this form element (may be null)
     */
    public Object getOrigVal();

    /**
     * Set the value for this element
     *
     * @param value the value for this element
     */
    public void setVal(Object ival);

    /**
     * Get the value for this element. If the underlying object
     * is actually an array (ie. allowMultiples = true), then
     * you should really be calling getVals() to get the whole
     * object array; if you call this method, you will just get 
     * the first element of the array.
     *
     * @return the value for this form element (may be null)
     */
    public Object getVal();
    
    /**
     * Get all the values for this element. There will always be an array returned.
     * if the value for this FormElement is null, an empty array is returned. This
     * method only really makes sense if allowMultiples = true
     *
     * @return the array of values for this form element
     */
    public Object[] getVals();

    /**
     * Get the value for this element as a String
     */
    public String getStringVal();

    /**
     * Get the value for this element as a String, defaulting
     * accordingly if the value is null
     */
    public String getStringVal(String dflt);

    /**
     * Get the value for this element as a Boolean
     */
    public Boolean getBooleanVal();

    /**
     * Get a Boolean value from the map, defaulting accordingly if
     * the value is null
     */
    public Boolean getBooleanVal(Boolean dflt);

    /**
     * Get the value for this element as a Integer
     */
    public Integer getIntegerVal();

    /**
     * Get a Integer value from the map, defaulting accordingly if
     * the value is null
     */
    public Integer getIntegerVal(Integer dflt);

    /**
     * Get the value for this element as a Date
     */
    public Date getDateVal();

    /**
     * Get a Date value from the map, defaulting accordingly if
     * the value is null
     */
    public Date getDateVal(Date dflt);

    /**
     * Get the value for this element as a Long
     */
    public Long getLongVal();

    /**
     * Get a Long value from the map, defaulting accordingly if
     * the value is null
     */
    public Long getLongVal(Long dflt);
    
    /**
     * Get the value for this element as a Short
     */
    public Short getShortVal();

    /**
     * Get a Short value from the map, defaulting accordingly if
     * the value is null
     */
    public Short getShortVal(Short dflt);

    /**
     * Get the value for this element as a Double
     */
    public Double getDoubleVal();

    /**
     * Get a Double value from the map, defaulting accordingly if
     * the value is null
     */
    public Double getDoubleVal(Double dflt);

    /**
     * Get the value for this element as a Float
     */
    public Float getFloatVal();

    /**
     * Get a Float value from the map, defaulting accordingly if
     * the value is null
     */
    public Float getFloatVal(Float dflt);

    /**
     * Set the parse exception associated with this element
     *
     * @return any parse exceptions associated with the element
     */
    public void setParseException(ParseException pe);

    /**
     * Get any parse exceptions associated with the element (ie. that
     * might have occurred when the element was mapped)
     *
     * @return any parse exceptions associated with the element
     */
    public ParseException getParseException();
    



//I don't think these are needed if we go back to the older/terser naming convention


    /**
     * Get the value for this element. If the elment has multiple values, the
     * first is returned.
     *
     * @return the value for this form element (may be null)
     */
//    public Object getSingleValue();
    

    /**
     * Get the values for this element. There will always be an array returned, even if
     * the value for this FormElement is null - then an empty array is returned.
     *
     * @return the array of values for this form element
     */

//    public Object [] getMultipleValues();

    /**
     * @clientCardinality 1 
     */
    /*#FormValidator lnkFormValidator;*/
}
