/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site (http://www.enhydra.org/).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: ValidationException.java,v 1.8 2003/05/17 15:57:09 christianc Exp $
 */
package org.enhydra.barracuda.core.forms;

import java.io.*;
import java.util.*;

import org.enhydra.barracuda.plankton.exceptions.NestableException;

/**
 * This class defines the validation exception. Code that catches
 * these exceptions should check for subExceptions, as this class
 * can be used to group collections of ValidationExceptions.
 */
public class ValidationException extends NestableException {

    protected List subExceptions = null;
    protected Object source = null;
    protected static final String sep = System.getProperty("line.separator");
    
    /**
     * The noargs public contructor for ValidationException
     */
    public ValidationException () {
        this("Generic Validation Exception");
    }

    /**
     * The public contructor for ValidationException
     *
     * @param s a String describing the exception
     */
    public ValidationException (String s) {
        this(null, s);
    }

    /**
     * The public contructor for ValidationException
     *
     * @param source the object which caused this error (usually a
     *        FormElement)
     */
    public ValidationException (Object source) {
        this(source, null);
    }

    /**
     * The public contructor for ValidationException
     *
     * @param source the object which caused this error (usually a
     *        FormElement)
     * @param s a String describing the exception
     */
    public ValidationException (Object source, String s) {
        this(source, s, null);
    }

    /**
     * The public contructor for ValidationException
     *
     * @param source the object which caused this error (usually a
     *        FormElement)
     * @param s a String describing the exception
     * @param iError the Error code associated with this exception
     */
    public ValidationException (Object isource, String s, Exception ibaseException) {
        super(s, ibaseException);
        source = isource;
    }

    /**
     * Get the form element which caused this error
     *
     * @return the form element which caused this error
     */
    public Object getSource() {
        return source;
    }
    
    /**
     * See if this particular ValidationException has
     * sub-exceptions
     *
     * @return true if this particular ValidationException has
     *         sub-exceptions
     */
    public boolean hasSubExceptions() {
        return (subExceptions!=null && subExceptions.size()>0);
    }
    
    /**
     * Add a sub exception
     *
     * @param ve a sub-exception to be added
     */
    public void addSubException(ValidationException ve) {
        if (subExceptions==null) subExceptions = new ArrayList();
        subExceptions.add(ve);
    }

    /**
     * Get a copy of the list of sub-exceptions. May be null
     * if there are no sub-exceptions.
     *
     * @return a copy of the list of sub-exceptions
     */
    public List getSubExceptions() {
        return (subExceptions==null ? null : new ArrayList(subExceptions));
    }
    
    public String toString() {
        StringBuffer sb = new StringBuffer(200);
        sb.append(this.getClass().getName()+"@"+Integer.toHexString(this.hashCode()));
        sb.append(", Msg:"+getMessage());
        if (source!=null && source instanceof ValidationException) sb.append(", Src: ValidationException[@"+Integer.toHexString(source.hashCode())+"]");
        else sb.append(", Src:"+source);
        if (subExceptions!=null) {
            sb.append(sep+"Sub-exceptions:");
            Iterator it = subExceptions.iterator();
            while (it.hasNext()) {
                sb.append(sep+"    "+it.next());
            }
        }
        return sb.toString();
    }

    /**
     * Collapse all sub-exceptions into a single list of
     * Validation exceptions
     */
    public List getExceptionList() {
        List list = new ArrayList();
        list.addAll(getExceptionList(this));
        return list;
    }
    
    protected List getExceptionList(ValidationException ve) {
        List list = new ArrayList();
        if (!ve.hasSubExceptions()) {
            list.add(ve);
        } else {
            Iterator it = ve.getSubExceptions().iterator();
            while (it.hasNext()) {
                list.addAll(getExceptionList((ValidationException) it.next()));
            }
        }
        return list;
    }

}
