/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site (http://www.enhydra.org/).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: AbstractPData.java,v 1.1 2002/11/02 19:15:53 cryd0221 Exp $
 */
package org.enhydra.barracuda.plankton.data;

import java.io.*;
import java.util.*;

/**
 * Abstract implementation of the basic PData methods.
 */
public abstract class AbstractPData implements PData {

    protected PData parent = null;
    protected boolean inheritParents = true;
    protected DefaultStateMap state = null;


    //--------------- PData --------------------------------------
    /**
     * set the object's parent. May be some other PData object or 
     * null (indicating hierarchy root). May not be a reference to self.
     *
     * @param p the parent PData object
     */
    public void setParent(PData p) {
        //don't ever allow an object to set itself as the parent...this
        //is meaningless and would result in an endless loop in the case 
        //of getRootParent()
        if (p==this) return;    
        parent = p;
    }

    /**
     * get the objects parent (null if root)
     *
     * @return the PData object
     */
    public PData getParent() {
        return parent;
    }

    /**
     * get the root parent by chaining back up the heirarchy until we find 
     * the highest PData object in the heirarchy.
     *
     * @return the root PData object
     */
    public PData getRootParent() {
        if ((parent!=null) || (parent instanceof PData)) return parent.getRootParent();
        else return this;
    }
    
    /**
     * Do we want to inherit parents. Defaults to true. This indicates that 
     * when an object is added to this data collection, if it implements
     * PData, it should automatically inherit the current object as its 
     * parent. You would typically set this to false if you were going
     * to be storing PData "mini-hierarchies" within a larger PData 
     * hierarchy
     * 
     * @param val true if we should inherit parents
     */
    public void setInheritParents(boolean val) {
        inheritParents = val;
    }
    
    /**
     * Return true if we are inheriting parents.
     * 
     * @return true if we are inheriting parents
     */
    public boolean isInheritParents() {
        return inheritParents;
    }


    //--------------- StateMap -----------------------------------
    /**
     * set a property in this StateMap
     *
     * @param key the key object
     * @param val the value object
     */    
    public void putState(Object key, Object val) {
        if (state==null) state = new DefaultStateMap();
        state.putState(key, val);
    }
    
    /**
     * get a property in this StateMap
     *
     * @param key the key object
     * @return the value for the given key
     */    
    public Object getState(Object key) {
        if (state==null) return null;
        else return state.getState(key);
    }
    
    /**
     * remove a property in this StateMap
     *
     * @param key the key object
     * @return the object which was removed
     */    
    public Object removeState(Object key) {
        if (state==null) return null;
        else return state.removeState(key);
    }
    
    /**
     * get a List of the keys for this StateMap (implementation
     * is an ArrayList)
     *
     * @return a List the keys for this StateMap
     */    
    public List getStateKeys() {
        if (state==null) return null;
        else return state.getStateKeys();
    }
    
    /**
     * get a copy of the underlying Map that holds 
     * the state values
     *
     * @return a copy of the underlying state Map
     */    
    public Map getStateValues() {
        if (state==null) state = new DefaultStateMap();
        return state.getStateValues();
    }




}
