/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site (http://www.enhydra.org/).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: ObjectRepository.java,v 1.2 2003/01/21 05:43:02 jacobk Exp $
 */
package org.enhydra.barracuda.plankton.data;

import java.util.*;
import javax.servlet.*;
import javax.servlet.http.*;

import org.apache.log4j.*;

/**
 * This class provides access to several different generic statemap repositories,
 * scoped for Global, Session, Local, or custom (NameSpace or Name...you provide 
 * the cleanup). The basic idea here is to make it easy to access object repositories
 * without having to pass references around.
 *
 * The Global object repository represents a common (threadsafe) statemap that is 
 * shared across the JVM. You would typically put things here like DataSources, global
 * variables, etc. Any objects that you make available globally need to be threadsafe.
 *
 * The Session object repository is a statemap wrapper around HttpSession. This provides
 * a convenient bridge to the Session interface, without having to have a reference to 
 * the HttpServletRequest. Storing items in this object places them in the underlying session.
 * Note that ApplicationGateway and ComponentGateway classes invoke the setupSessionRepository()
 * call for you, so all you have to do is call getSessionRepository() and you're in business.
 *
 * The Local object repository is a statemap object that lasts for the duration of a req-resp
 * cycle. Note that the ApplicationGateway and ComponentGateway classes clean up both Session and
 * Local repositories.
 *
 * You can also get non-scoped ObjectRepositories by using either NameSpace or String keys. IF you
 * use this approach, its up to you to manually remove the repositories when you're done with them
 * (or else you'll end up leaving them in memory taking up space).
 *
 * @author Christian Cryder
 * @version 1.0
 */
public class ObjectRepository extends DefaultStateMap {

    //public constants
    protected static final Logger logger = Logger.getLogger(ObjectRepository.class.getName());

    protected static ObjectRepository global = new ThreadsafeRepository("GlobalOR");
    protected static ObjectRepository weakGlobal = new WeakThreadsafeRepository("WeakGlobalOR");    //csc_082302.2 - added
    protected static Map session = new HashMap();
    protected static Map weaksession = new HashMap();
    protected static Map local = new HashMap();
    protected static Map custom = new HashMap();

    protected String name = "[unnamed]";

    /**
     * protected constructor
     */
    protected ObjectRepository() {
        this(null);
    }
    protected ObjectRepository(String iname) {
        if (iname!=null) name = iname;
    }
    
    /**
     * Get a reference to the Global repository. This repository
     * is shared across the entire JVM (consequently its threadsafe). 
     * The global repository never really goes away, so you don't need 
     * to do any cleanup on this one.
     */
    public static ObjectRepository getGlobalRepository() {
        return global;
    }

    //csc_082302.2 - added
    /**
     * Get a reference to the Weak Global repository. This repository
     * is shared across the entire JVM (consequently its threadsafe),
     * but like the Weak Session repository, its backed by a weak hash
     * map, allowing object which are placed in here to be gc'd as needed.
     */
    public static ObjectRepository getWeakGlobalRepository() {
        return weakGlobal;
    }

    /**
     * Set up the Session repository for this particular thread. You must invoke this 
     * method before actually getting the session repository, IFF you actually
     * want the repository to wrapp the Servlet's Session structure. Otherwise, 
     * when you get the Session repository (below) it will simply return a regular
     * state map.
     */
    public static void setupSessionRepository(HttpServletRequest req) {
        String key = "SessionOR_"+Thread.currentThread().getName();
        if (logger.isInfoEnabled()) logger.info("Setting up session repository: "+key);
//System.out.println("Setting up session repository: "+key);
        synchronized (session) {
            session.put(key, new SessionRepository(req));
        }
    }

    /**
     * Get a reference to the Session repository. If you are using this 
     * within Barracuda (ie. ApplicationGateway or ComponentGateway), this
     * will return a StateMap that wraps the Servlet session.
     */
    public static ObjectRepository getSessionRepository() {
        String key = "SessionOR_"+Thread.currentThread().getName();
        if (logger.isInfoEnabled()) logger.info("Getting session repository: "+key);
//System.out.println("Getting session repository: "+key);
        ObjectRepository or = (ObjectRepository) session.get(key);
        if (or==null) {
            if (logger.isInfoEnabled()) logger.info("Setting up default session respository: "+key);
//System.out.println("Setting up default session respository: "+key);
            or = new ObjectRepository(key);
            synchronized (session) {
                session.put(key, or);
            }
        }
        return or;    
    }

    /**
     * Release the Session repository
     */
    public static void removeSessionRepository() {
        String key = "SessionOR_"+Thread.currentThread().getName();
        synchronized (session) {
            session.remove(key);
        }
        if (logger.isInfoEnabled()) logger.info("Removed session repository: "+key);
//System.out.println("Removed session repository: "+key);
    }

    //csc_030602.1 - added
    /**
     * Get a reference to a Weak Session repository. Objects placed in here
     * are scoped to Session, but may be reclaimed by garbage collecter as needed
     * (so you must always check for null and reinitialize if necessary)
     */
    public static ObjectRepository getWeakSessionRepository() {
        String key = "WeakSessionOR_$$";
        if (logger.isInfoEnabled()) logger.info("Getting weak session repository: "+key);
        ObjectRepository session_or = getSessionRepository();
//System.out.println("Getting session repository: "+key);
        ObjectRepository weak_or = (ObjectRepository) session_or.getState(key);
        if (weak_or==null) {
            if (logger.isInfoEnabled()) logger.info("Setting up weak session respository: "+key);
//System.out.println("Setting up default session respository: "+key);
            weak_or = new WeakRepository(key);
            synchronized (session_or) {
                session_or.putState(key, weak_or);
            }
        }
        return weak_or;    
    }

    /**
     * Get a reference to the Local repository. This repository is shared
     * across the thread. If you use this repository, it is your responsibility 
     * to clean things up (unless something else is cleaning it up for you,
     * like the Barracuda ApplicationGateway or ComponentGateway)
     */
    public static ObjectRepository getLocalRepository() {
        String key = "LocalOR_"+Thread.currentThread().getName();
        if (logger.isInfoEnabled()) logger.info("Getting local repository: "+key);
        ObjectRepository or = (ObjectRepository) local.get(key);
        if (or==null) {
            if (logger.isInfoEnabled()) logger.info("Creating new local repository: "+key);
            or = new ObjectRepository(key);
            local.put(key, or);
        }
        return or;
    }

    /**
     * Release the Local repository.
     */
    public static void removeLocalRepository() {
        String key = "LocalOR_"+Thread.currentThread().getName();
        local.remove(key);
        if (logger.isInfoEnabled()) logger.info("Removed local repository: "+key);
    }

    /**
     * Get an object repository associated with a given NameSpace. If
     * the object repository for this namespace does not already exist,
     * it will be created automatically.
     */
    public static ObjectRepository getObjectRepository(NameSpace ns) {
        if (logger.isInfoEnabled()) logger.info("Getting custom repository: "+ns);
        ObjectRepository or = (ObjectRepository) custom.get(ns);
        if (or==null) {
            if (logger.isInfoEnabled()) logger.info("Creating new custom repository: "+ns);
            or = new ObjectRepository(ns.getName());
            synchronized (custom) {
                custom.put(ns, or);
            }
        }
        return or;
    }
    
    /**
     * Release an object repository associated with a given NameSpace
     */
    public static void removeObjectRepository(NameSpace ns) {
        synchronized (custom) {
            custom.remove(ns);
        }
        if (logger.isInfoEnabled()) logger.info("Removed custom repository: "+ns);
    }

    /**
     * Get an object repository based on a given name. If
     * the object repository for this namespace does not already exist,
     * it will be created automatically.
     */
    public static ObjectRepository getObjectRepository(String name) {
        if (logger.isInfoEnabled()) logger.info("Getting custom repository: "+name);
        ObjectRepository or = (ObjectRepository) custom.get(name);
        if (or==null) {
            if (logger.isInfoEnabled()) logger.info("Creating new custom repository: "+name);
            or = new ObjectRepository(name);
            synchronized (custom) {
                custom.put(name, or);
            }
        }
        return or;
    }
    
    /**
     * Release an object repository associated with a given name
     */
    public static void removeObjectRepository(String name) {
        synchronized (custom) {
            custom.remove(name);
        }
        if (logger.isInfoEnabled()) logger.info("Removed custom repository: "+name);
    }

    /**
     * Return the name of this object repository
     */
    public String getName() {
        return name;
    }

    /**
     * A threadsafe wrapper around ObjectRepository
     */
    static class ThreadsafeRepository extends ObjectRepository {
        public ThreadsafeRepository() {super();}
        public ThreadsafeRepository(String iname) {super(iname);}
        
        public synchronized void putState(Object key, Object val) {
            super.putState(key, val);
        }
        public synchronized Object getState(Object key) {
            return super.getState(key);
        }
        public synchronized Object removeState(Object key) {
            return super.removeState(key);
        }
        public synchronized List getStateKeys() {
            return super.getStateKeys();
        }
        public synchronized Map getStateValues() {
            return super.getStateValues();
        }
    }

    //csc_030602.1 - added
    /**
     * A weak wrapper around ObjectRepository (things that get put in 
     * here can get reclaimed by the garbage collecter as needed, so
     * you must alwasy check for null as you retrieve objects from this 
     * space)
     */
    static class WeakRepository extends ObjectRepository {
        Map weakMap = new WeakHashMap();
    
        public WeakRepository() {super();}
        public WeakRepository(String iname) {super(iname);}
        
        public void putState(Object key, Object val) {
            weakMap.put(key, val);
        }
        public Object getState(Object key) {
            return weakMap.get(key);
        }
        public Object removeState(Object key) {
            return weakMap.remove(key);
        }
        public List getStateKeys() {
            return new ArrayList(weakMap.keySet());
        }
        public Map getStateValues() {
            return new HashMap(weakMap);
        }
    }

    //csc_082302.2 - added
    /**
     * A weak threadsafe wrapper around ObjectRepository
     */
    static class WeakThreadsafeRepository extends WeakRepository {
        public WeakThreadsafeRepository() {super();}
        public WeakThreadsafeRepository(String iname) {super(iname);}
        
        public synchronized void putState(Object key, Object val) {
            super.putState(key, val);
        }
        public synchronized Object getState(Object key) {
            return super.getState(key);
        }
        public synchronized Object removeState(Object key) {
            return super.removeState(key);
        }
        public synchronized List getStateKeys() {
            return super.getStateKeys();
        }
        public synchronized Map getStateValues() {
            return super.getStateValues();
        }
    }

    /**
     * The basic idea behind this is that we want to wrap the session
     * as a ObjectRepository object. In order to make this work, it's important
     * that we don't actually cause the session to be instantiated
     * unless absolutely necessary
     */    
    static class SessionRepository extends ObjectRepository {
        HttpServletRequest req = null;
        StateMap map = null;

        public SessionRepository() {super();}
        public SessionRepository(String iname) {super(iname);}

        public SessionRepository(HttpServletRequest ireq) {
            req = ireq;
        }
        public synchronized void putState(Object key, Object val) {
            if (map==null) {
                HttpSession session = req.getSession();
                map = new HttpSessionStateMap(session);
            }
            map.putState(key, val);
        }
        public synchronized Object getState(Object key) {
            if (map==null) {
                HttpSession session = req.getSession(false);
                if (session==null) return null;
                else map = new HttpSessionStateMap(session);
            }
            return map.getState(key);
        }
        public synchronized Object removeState(Object key) {
            if (map==null) {
                HttpSession session = req.getSession(false);
                if (session==null) return null;
                else map = new HttpSessionStateMap(session);
            }
            return map.removeState(key);
        }
        public synchronized List getStateKeys() {
            if (map==null) {
                HttpSession session = req.getSession(false);
                if (session==null) return new ArrayList();
                else map = new HttpSessionStateMap(session);
            }
            return map.getStateKeys();
        }
        public synchronized Map getStateValues() {
            if (map==null) {
                HttpSession session = req.getSession(false);
                if (session==null) return new HashMap();
                else map = new HttpSessionStateMap(session);
            }
            return map.getStateValues();
        }
    }

}