/*
 * Enhydra Java Application Server Project
 *
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site (http://www.enhydra.org/).
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific terms governing rights and limitations
 * under the License.
 *
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 *
 * Contributor(s):
 *  @author christianc@barracudamvc.org
 *
 * $Id: Localize.java,v 1.28 2003/05/14 16:28:35 jacobk Exp $
 */
package org.enhydra.barracuda.taskdefs;

import java.util.*;
import java.io.*;
import org.apache.tools.ant.*;
import org.apache.tools.ant.taskdefs.*;
import org.apache.tools.ant.types.*;
import org.enhydra.xml.xmlc.taskdef.*;

//csc_112801 - created this class
/**
 * Ant task to invoke the localize HTML template files. Supports all the
 * attributes found in the Xmlc taskdef. This is a reworking of the original
 * Localize taskdef--the taskdef reads the master templates, generates locale
 * templates from it using straight text substitution (previously we actually
 * compiled the master template via XMLC and then parsed the resulting DOM
 * object, which caused all kinds of problems because of Tidy bugs/quirks).
 * At any rate, once the templates are generated, then we invoke XMLC on
 * everything, and then at the end we compile it all using the javac taskdef.
 *
 * There is probably room to improve the performance of this thing (I haven't
 * done any optimization)
 *
 * @author christianc@barracudamvc.org
 */
public class Localize extends Xmlc {

    private static String sep = System.getProperty("line.separator");

    static {
        // if not already has a xmlc utils class installed
        // then install the localize specific version
        if (null==System.getProperty(XmlcUtils.DEFAULT_UTILS_CLASS_KEY)) {
            System.setProperty(XmlcUtils.DEFAULT_UTILS_CLASS_KEY, LocalizeXmlcUtilsImpl.class.getName());
        }
    }

//    protected static String logHdr1 = " [localize_dbg] ";
    protected String classpathRef = null;
    protected boolean tidy = true;

    //jdv_121201 - these two Strings represent the file name and location of a file
    //used for master localization.  if they remain null, localization functions normally
    public String masterFile=null; //the name of the master file for localization
    public String masterDir=null;  //the location of the file for localization

    public final static String[] countryCodes = new String[] {
        "aa", "ab", "af", "am", "ar", "as", "ay", "az", "ba", "be", "bg", "bh",
        "bi", "bn", "bo", "br", "ca", "co", "cs", "cy", "da", "de", "dz", "el",
        "en", "eo", "es", "et", "eu", "fa", "fi", "fj", "fo", "fr", "fy", "ga",
        "gd", "gl", "gn", "gu", "ha", "he", "hi", "hr", "hu", "hy", "ia", "id",
        "ie", "ik", "is", "it", "iu", "ja", "jw", "ka", "kk", "kl", "km", "kn",
        "ko", "ks", "ku", "ky", "la", "ln", "lo", "lt", "lv", "mg", "mi", "mk",
        "ml", "mn", "mo", "mr", "ms", "mt", "my", "na", "ne", "nl", "no", "oc",
        "om", "or", "pa", "pl", "ps", "pt", "qu", "rm", "rn", "ro", "ru", "rw",
        "sa", "sd", "sg", "sh", "si", "sk", "sl", "sm", "sn", "so", "sq", "sr",
        "ss", "st", "su", "sv", "sw", "ta", "te", "tg", "th", "ti", "tk", "tl",
        "tn", "to", "tr", "ts", "tt", "tw", "ug", "uk", "ur", "uz", "vi", "vo",
        "wo", "xh", "yi", "yo", "za", "zh", "zu"
        };

    /**
     * Set the classpath ref for javac.
     */
    public void setClasspathref(String iclasspathRef) {
        classpathRef = iclasspathRef;
    }

    /**
     * Run tidy on the localized templates
     * @param tidy Set "true" for tidy
     */
    public void setTidy(BooleanAttribute itidy) {
        tidy = (itidy.getValue().equals("yes") || itidy.getValue().equals("true"));
    }

    //jdv_121201
    /**
     * Set the Master Locale Directory by accesssing the build.xml file
     * @param masterDir the location of the Master Localization File
     */
    public void setMasterlocaledir(String masterDir) {
        this.masterDir=masterDir;
//System.out.println("MASTER DIR " +masterDir);
    }

    //jdv_121201
    /**
     * Set the Master Locale File Name by accessing the build.xml file
     * @param masterFile the given name of the Master Localization File
     */
    public void setMasterlocalefile(String masterFile) {
        this.masterFile=masterFile;
//System.out.println("MASTER FILE " +masterFile);
    }

    /**
     * Execute the task. The basic flow here is: invoke xmlc to generate
     * template classes, invoke the java compiler to compile those classes,
     * then use those compiled DOM structures to create localized versions of
     * the XML templates. Finally, reinvoke xmlc and javac to recompile the
     * newly created localized templates.
     */
    public void execute() throws BuildException {
        //start by adjusting the log level
        if (verbose) {
            logLevel = project.MSG_INFO;
        } else {
            logLevel = project.MSG_DEBUG;
        }

        //now we can create localized templates
        log("Searching for XMLC Localization templates in "+srcDir+"...", logLevel);

        // now scan source dir
        boolean wroteFiles = false;
        boolean touchedTemplate = false;
        int fileCount = 0;
        int templateCount = 0;
        DirectoryScanner ds = this.getDirectoryScanner(srcDir);
        String[] files = ds.getIncludedFiles();
        iterateFiles: for (int i=0; i<files.length; i++) {
            //read our file list
            String file = files[i];
            String fileName = file;
            String fileExt = null;
            int epos = file.lastIndexOf(".");
            if (epos>-1) {
                fileName = file.substring(0,epos);
                if (epos+1<file.length()) fileExt = file.substring(epos+1);
            }
            File templateFile = new File(srcDir, files[i]);
            long templateLastTouched = templateFile.lastModified();

            //first and foremost, figure out if we're looking at an
            //existing localized template file. If so, skip it.
            if (fileName.indexOf("_")>0) {
                String tfile = fileName;
                if (!tfile.endsWith("_")) tfile += "_";
                StringTokenizer st = new StringTokenizer(tfile, "_");
                while (st.hasMoreTokens()) {
                    String token = st.nextToken();
                    int pos = Arrays.binarySearch(countryCodes, token);
                    if (pos>-1 && pos<countryCodes.length && countryCodes[pos].equals(token)) continue iterateFiles;
                }
            }

            //figure out if the template has a properties file; if the props file is
            //newer than the template, touch the template

            //jdv_121201_start - if the masterFile!=null, use it instead of the
            //regular propFile
//            File propFile = new File(srcDir, fileName+".properties");
            File propFile = null;
            if (masterFile!=null) {
                propFile = new File(masterDir, masterFile+".properties");
            } else {
                propFile = new File(srcDir, fileName+".properties");
            }
            //jdv121201_end

            long propLastTouched = -1;
            if (!propFile.exists()) propLastTouched = propFile.lastModified();
            if (propLastTouched >= templateLastTouched) {
                if (!touchedTemplate) {
                    //touch the template so it'll get recompiled
                    templateFile.setLastModified(System.currentTimeMillis());
                    touchedTemplate = true;
//System.out.println ("Touching template at pt 2");
                }
            }

            //see if the master template has a .xmlc file. If not, create it.
            File xmlcFile = new File(srcDir, fileName+".xmlc");
            createMasterXMLC: if (propFile.exists() && !xmlcFile.exists()) {
                //create the new .xmlc file
                BufferedWriter writer = null;
                try {
                    //write the contents out
                    writer = new BufferedWriter(new FileWriter(xmlcFile));
                    writer.write("-generate both");
                    writer.newLine();
                    writer.flush();

                    if (!touchedTemplate) {
                        //touch the template so it'll get recompiled
                        templateFile.setLastModified(System.currentTimeMillis());
                        touchedTemplate = true;
//System.out.println ("Touching template at pt 1");
                    }
                } catch (IOException e) {
                    if (writer!=null) try {writer.close();} catch (IOException ioe) {}
                }
            }

            //make sure there's a .class file; if not, we know we need to recompile
            String baseClassFileName = "";
            if (packageDir!=null) baseClassFileName = packageDir + "/";
            baseClassFileName += fileName + ML + ".class";
            File baseClassFile = new File(destDir, baseClassFileName);
            long baseClassLastTouched = -1;
            if (!baseClassFile.exists()) {
                if (!touchedTemplate) {
                    //touch the template so it'll get recompiled
                    templateFile.setLastModified(System.currentTimeMillis());
                    touchedTemplate = true;
//System.out.println ("Touching template at pt 3");
                }
            } else {
                baseClassLastTouched = baseClassFile.lastModified();
                if (forceBuild ||
                    templateLastTouched >= baseClassLastTouched ||
                    propLastTouched >= baseClassLastTouched) {
                    if (!touchedTemplate) {
                        //touch the template so it'll get recompiled
                        templateFile.setLastModified(System.currentTimeMillis());
                        touchedTemplate = true;
//System.out.println ("Touching template at pt 4");
//System.out.println ("templateLastTouched:  "+templateLastTouched);
//System.out.println ("propLastTouched:      "+propLastTouched);
//System.out.println ("baseClassLastTouched: "+baseClassLastTouched);
                    }
                }
            }
            templateLastTouched = templateFile.lastModified();
//System.out.println ("baseClassFile: "+baseClassFile);

            //now if there is no props file, then there is no point continuing
            //at this point, since everything below this relates to localization
            if (!propFile.exists()) continue;

            //now see if there are other properties files with different locale info
            //(the purpose of this loop is to figure out which templates need to
            //be rebuilt and to load the appropriate prop files for each)
            DirectoryScanner dsLocales = new DirectoryScanner();
            //jdv_121201_start
//            DirectoryScanner dsLocales = new DirectoryScanner();
//            String[] includes = {fileName+"*.properties"};
//            dsLocales.setIncludes(includes);
//            dsLocales.setBasedir(srcDir);
//            dsLocales.scan();
//            String[] localeFiles = dsLocales.getIncludedFiles();
//            List locales = new ArrayList();
//            HashMap propBundleMap = new HashMap();
//            int spos = fileName.length();

            //if the masterFile!=null, the spos is the length of the masterFile,
            //the baseDir is set to masterDir, and the filename to scan for is the
            //masterFile. otherwise, the standard format is used.
            int spos = -1;
            String includeFileName;
            if (masterFile!=null) {
               includeFileName = masterFile;
               dsLocales.setBasedir(masterDir);
               spos = masterFile.length();
            } else {
               includeFileName = fileName;
               dsLocales.setBasedir(srcDir);
               spos = fileName.length();
            }
            String[] includes = {includeFileName+"*.properties"};
            dsLocales.setIncludes(includes);
            dsLocales.scan();
            String[] localeFiles = dsLocales.getIncludedFiles();
            List locales = new ArrayList();
            HashMap propBundleMap = new HashMap();
            //jdv_121201_end

            iteratePropFiles:  for (int j=0; j<localeFiles.length; j++) {
                //make sure we're loking at a localization prop file
                String localeFile = localeFiles[j];
                epos = localeFile.length()-11;
                String localeSubstr = localeFile.substring(spos,epos);
                if (!(localeSubstr.startsWith("_"))) continue;

                // get the html-equivalent of the localization prop file
                //jdv_053102_start
                //Use the fileName + localeSubstr to reference the proper file no
                //matter what the localeFile is.
                //int dotPos = localeFile.indexOf(".");
                //String htmlFile = localeFile.substring(0, dotPos) + "." + fileExt;
                String htmlFile = fileName + localeSubstr + "." + fileExt;
                //jdv_053102_end
                File localizedHtmlFile = new File(srcDir, htmlFile);
                //jdv_121201_start - if masterFile!=null, replace the srcDir with masterDir
//                File localizedPropFile = new File(srcDir, localeFile);
                File localizedPropFile = null;
                if (masterFile!=null) {
                    localizedPropFile = new File(masterDir, localeFile);
                } else {
                    localizedPropFile = new File(srcDir, localeFile);
                }
                //jdv_121201_end

                //see if the generated template has an .xmlc file. If not, create it.
                //jdv_053102_start
                //As Above, use the fileName and localeSubstring to make sure the proper
                //file is referenced.
                //String xmlcOptionsFileName = localeFile.substring(0, dotPos)+".xmlc";
                String xmlcOptionsFileName = fileName + localeSubstr + ".xmlc";
                //jdv_053102_end
                File xmlcOptionsFile = new File(srcDir, xmlcOptionsFileName);
                boolean autogeneratedOptionsFile = false;
                createMasterXMLC: if (!xmlcOptionsFile.exists()) {

                    //figure out parent class name
                    //mr_022602_start
                    // String clStr = baseClassFileName;
                    // int epos1 = clStr.indexOf(".class");
                    // clStr = clStr.substring(0,epos1);
                    XmlcUtils xmlcUtils = XmlcUtils.create();
                    String clStr = xmlcUtils.buildFullBaseFileName(packageName, packageDir, baseClassFileName);
                    //mr_022602_end
                    clStr = clStr.replace('\\','.');
                    clStr = clStr.replace('/','.');

                    //create the new .xmlc file
                    BufferedWriter writer = null;
                    try {
                        //write the contents out
                        writer = new BufferedWriter(new FileWriter(xmlcOptionsFile));
                        writer.write("-implements "+clStr);
                        writer.newLine();
                        writer.flush();
                        autogeneratedOptionsFile = true;
                        wroteFiles = true;
                    } catch (IOException e) {
                        if (writer!=null) try {writer.close();} catch (IOException ioe) {}
                    }
                }

                //see if we need to rebuild the template
                if (localizedHtmlFile.exists()) {
                    File baseHtmlFile = new File(srcDir, fileName + "." + fileExt);

                    long localizedHtmlLastTouched = localizedHtmlFile.lastModified();
                    long outputPropLastTouched = localizedPropFile.lastModified();

                    // rebuild if:
                    // 1) forceBuild is true, or
                    // 2) the template was touched
                    // 3) the base properties file is more recent than the locale html file, or
                    // 4) the locale properties file is more recent than the locale html file
                    boolean rebuild = false;
                    if (forceBuild ||
                        touchedTemplate ||
                        autogeneratedOptionsFile ||
                        propLastTouched >= localizedHtmlLastTouched ||
//                        baseClassLastTouched >= localizedHtmlLastTouched ||
                        outputPropLastTouched >= localizedHtmlLastTouched) {
                            rebuild = true;
//System.out.println ("Rebuild template at pt 1: "+localizedHtmlFile);
//System.out.println ("forceBuild:               "+forceBuild);
//System.out.println ("touchedTemplate:          "+touchedTemplate);
//System.out.println ("autogeneratedOptionsFile: "+autogeneratedOptionsFile);
//System.out.println ("propLastTouched:          "+propLastTouched);
//System.out.println ("outputPropLastTouched:    "+outputPropLastTouched);
//System.out.println ("localizedHtmlLastTouched: "+localizedHtmlLastTouched);
                    }

                    // if the base html file has a .xmlc file associated with it, factor
                    // that into the rebuild logic
                    if (!forceBuild) {
                        String baseXmlcFileName = fileName + ".xmlc";
                        File baseXmlcFile = new File(srcDir, baseXmlcFileName);
                        if (baseXmlcFile.exists() && baseXmlcFile.canRead())  {
                            // rebuild if
                            // 5) the base .xmlc file is more recent than the locale html file
                            long baseXmlcLastTouched = baseXmlcFile.lastModified();

                            //jdv_060602
                            // Should be only rebuild if the xmlc file is newer
                            // than the localized html file.
                            //if (baseXmlcLastTouched >= localizedHtmlLastTouched) {
                            if (baseXmlcLastTouched > localizedHtmlLastTouched) {
                                System.out.println(" REBUILD BECAUSE " + baseXmlcFile + " > " + localizedHtmlFile);
                                rebuild = true;
//System.out.println ("Rebuild template at pt 2: "+localizedHtmlFile);
                            }
                        }
                    }

                    // if the locale html file has a .xmlc file associated with it, factor that
                    // into the rebuild logic
                    if (!forceBuild) {
                        //jdv_053102_start
                        //Use the fileName + localeSubstr here to ensure that the file referenced is
                        //the same as the fileName, not the localeFile
                        //String localeXmlcFileName = localeFile.substring(0, dotPos) + ".xmlc";
                        String localeXmlcFileName = fileName + localeSubstr + ".xmlc";
                        //jdv_053102_end
                        File localeXmlcFile = new File(srcDir, localeXmlcFileName);
                        if (localeXmlcFile.exists() && localeXmlcFile.canRead())  {
                            // rebuild if
                            // 6) the locale-specific
                            long localeXmlcLastTouched = localeXmlcFile.lastModified();

                            //jdv_060602
                            // Should be only rebuild if the locale xmlc file is newer
                            // than the localized html file.
                            //if (localeXmlcLastTouched >= localizedHtmlLastTouched) {
                            if (localeXmlcLastTouched > localizedHtmlLastTouched) {
                                rebuild = true;
//System.out.println ("Rebuild template at pt 3: "+localizedHtmlFile);
                            }
                        }
                    }

                    if (!rebuild) {
                        log("        Template up-to-date: "+localizedHtmlFile.getName(), logLevel);
                        continue iteratePropFiles;
                    }
                }

                //figure out the language, country, and variant
                StringTokenizer st = new StringTokenizer(localeSubstr, "_");
                String language = "";
                String country = "";
                String variant = null;
                if (st.hasMoreTokens()) language = st.nextToken();
                if (st.hasMoreTokens()) country = st.nextToken();
                if (st.hasMoreTokens()) {
                    String sep = "";
                    while (st.hasMoreTokens()) {
                        if (variant==null) variant = st.nextToken();
                        else variant += st.nextToken();
                        sep = "_";
                    }
                }

                //figure out the targetLocale
                Locale targetLocale = null;
                if (variant!=null) targetLocale = new Locale(language, country, variant);
                else targetLocale = new Locale(language, country);
                locales.add(targetLocale);

                //find the actual prop file and place it in the propFileMap
                //jdv_121201_start - if the masterFile!=null, replace the
                //localePropFile with the masterFile and masterDir
//              File localePropFile = new File(srcDir, fileName+localeSubstr+".properties");
                File localePropFile = null;
                if (masterFile!=null) {
                    localePropFile = new File(masterDir, masterFile+localeSubstr+".properties");
                } else {
                    localePropFile = new File(srcDir, fileName+localeSubstr+".properties");
                }
                //jdv_121201_end
                if (!localePropFile.exists()) {
                    log("        Unable to read localization properties file "+localePropFile, project.MSG_WARN);
                    continue;
                }
                PropertyResourceBundle rb = null;
                try {rb = new PropertyResourceBundle(new FileInputStream(localePropFile));}
                catch (Exception e) {
                    log("        Unable to create resource bundle for file "+localePropFile, project.MSG_WARN);
                    continue;
                }
                propBundleMap.put(localeSubstr, rb);
            } //end iteratePropFiles

            // if there are no locales to work with, continue to the next html file
            if (locales.size()==0) continue;

            //read in the master template
            log("...Generating localized templates");
            StringBuffer sb = new StringBuffer(5000);
            try {
                FileReader fr = new FileReader(templateFile);
                BufferedReader br = new BufferedReader(fr);
                while (true) {
                    String str = br.readLine();
                    if (str==null) break;
                    sb.append(str);
                    sb.append(sep);
                }
                br.close();
            } catch (IOException e) {
                System.err.println ("Error reading template: "+e);
                continue;
            }
            String masterTemplateStr = sb.toString();
            log("    Loaded template "+file, logLevel);
            fileCount++;

            //read in the master properties file (note that we read this directly,
            //as opposed to using the Properties class, since its important we
            //process keys in the order that they're listed in the file. This will
            //allow more specific phrases to be translated first, and more general
            //ones (subsets of the more specific, perhaps) to be translated later)
            List masterKeyList = new ArrayList(100);
            List masterValList = new ArrayList(100);
            try {
                FileReader fr = new FileReader(propFile);
                BufferedReader br = new BufferedReader(fr);
                while (true) {
                    //read a line from the properties file
                    String str = br.readLine();
                    if (str==null) break;
                    str = str.trim();

                    //skip comment lines
                    if (str.startsWith("#")) continue;

                    //skip lines without an '='
                    int eqpos = str.indexOf('=');
                    if (eqpos<0) continue;

                    //find the keys and vals
                    String key = (""+str.substring(0,eqpos)).trim();
                    String val = (""+str.substring(eqpos+1)).trim();
                    masterKeyList.add(key);
                    masterValList.add(val);
                }
                br.close();
            } catch (IOException e) {
                System.err.println ("Error reading template: "+e);
                continue;
            }

             //now process each target locale
            Object locArray[] = locales.toArray();
            iterateTemplates: for (int j=0; j<locArray.length; j++) {
                //figure out the locale extension
                List locList = new ArrayList();
                Locale loc = (Locale) locArray[j];
                String locExt = "";
                String buf = loc.getLanguage();
                if (buf!=null && buf.length()>0) {
                    locExt+="_"+buf;
                    locList.add(0,propBundleMap.get(locExt));
                    buf = loc.getCountry();
                    if (buf!=null && buf.length()>0) {
                        locExt+="_"+buf;
                        locList.add(0,propBundleMap.get(locExt));
                        buf = loc.getVariant();
                        if (buf!=null && buf.length()>0) {
                            locExt+="_"+buf;
                            locList.add(0,propBundleMap.get(locExt));
                        }
                    }
                }

                //now we're ready to localize...
                //...first, create a new string from the master template
                String templateStr = new String(masterTemplateStr);

                //...now run through the key/val pairs
                Map lmap = new HashMap(100);
                for (int k=0, max=masterKeyList.size(); k<max; k++) {
                    //...now find the non-localized value for this key
                    String key = (String) masterKeyList.get(k);
                    String val = (String) masterValList.get(k);

                    //...now find the localized value for the key
                    String lczval = null;
                    Iterator it = locList.iterator();
                    iterate_rbs: while (it.hasNext()) {
                        ResourceBundle rb = (ResourceBundle) it.next();
                        try {lczval = rb.getString(key);}
                        catch (MissingResourceException e) {}
                        if (lczval!=null) break iterate_rbs;
                    }
                    if (lczval==null) continue;

                    //...figure out a string representation of the value's hashcode,
                    //...and substitute this into the templateStr (this will prevent
                    //...already localized text from getting replaced on subsequent
                    //...replaces)
                    String hash = "@"+Integer.toHexString(lczval.hashCode());
                    templateStr = qualifiedStrReplace(templateStr, val, hash);
                    lmap.put(hash, lczval);
                }

                //...once we've completed the replacing, iterate back through the
                //...localized value hash map and replace all the hashes with the
                //...real localized values
                Iterator it = lmap.keySet().iterator();
                while (it.hasNext()) {
                    String hashkey = (String) it.next();
                    String lczval = (String) lmap.get(hashkey);
                    templateStr = strReplace(templateStr, hashkey, lczval);
                }

                //...the last step is to simply write out the new template
                try {
                    File outputFile = new File(srcDir, fileName+locExt+"."+fileExt);
                    FileWriter fw = new FileWriter(outputFile);
                    BufferedWriter bw = new BufferedWriter(fw);
                    bw.write(templateStr);
                    bw.close();
                    wroteFiles = true;
                    templateCount++;
                } catch (IOException e) {
                    System.err.println ("Error writing localized template: "+e);
                    continue;
                }
            } //end iterateTemplates
        } //end iterateFiles

        if (touchedTemplate || wroteFiles) {
            //now that all the templates have been generated, go ahead and invoke
            //the XMLC processing
            log("...Invoking XMLC precompiler taskdef");
            int origLogLevel = logLevel;
            logLevel = project.MSG_DEBUG;
            super.execute();
            logLevel = origLogLevel;
        }

        //finally, invoke the compiler
        if (touchedTemplate || wroteFiles) log("...Invoking Javac compiler taskdef");
        InternalJavac compile = new InternalJavac();
        compile.setProject(project);
        compile.setSrcdir(new Path(project, srcOutDir.getPath()));
        compile.setDestdir(destDir);
        if (classpathRef!=null) compile.setClasspathRef(new Reference(classpathRef));
        compile.execute();

        //hey hey! we're done!
        log("Finished!", logLevel);
        if (fileCount!=0 || templateCount!=0 || verbose) log("Processed "+files.length+" source files, found "+fileCount+" master templates, created "+templateCount+" new localized templates"+(fileCount>=0 && templateCount==0 ? " (all files up-to-date)" : ""));
    }

    /**
     * Replace occurrences of str1 in string str with str2, but only
     * if the str1 is qualified (ie. not embedded in an 'id=""' or
     * 'class=""' attribute)
     */
    protected String qualifiedStrReplace(String str, String str1, String str2) {
        int spos = 0;
        while (true) {
            //search for a match
            int found = str.indexOf(str1, spos);
            if (found<0) break;

            // write the replacement str2
            if (str2!=null && isQualified(str, found)) {
                str = str.substring(0,found)+str2+str.substring(found+str1.length());
                spos = found + str2.length();
            } else {
                spos = found + str1.length();
            }
        }
        return str;
    }

    /**
     * Replace occurrences of str1 in string str with str2
     */
    protected String strReplace(String str, String str1, String str2) {
        StringBuffer ret = new StringBuffer();
        int start = 0;
        int found = str.indexOf(str1);
        while (found >= 0) {
            // write everything up to the found str1
            if (found > start) {
                ret.append(str.substring(start, found));
            }

            // write the replacement str2
            if (str2!=null) {
                ret.append(str2);
            }

            // search again
            start = found + str1.length();
            found = str.indexOf(str1,start);
        }

        // write the remaining characters
        if (str.length() > start) {
            ret.append(str.substring(start, str.length()));
        }

        return ret.toString();
    }

    protected boolean isQualified(String str, int fpos) {
        if (embeddedInToken(str, fpos, "id=\"","\"")) return false;
        if (embeddedInToken(str, fpos, "class=\"","\"")) return false;
        return true;
    }

    protected boolean embeddedInToken(String str, int fpos, String stoken, String etoken) {
        //check for the first occurrence of the start token prior to fpos; if
        //its missing, then we can obviously return false
        int spos = str.lastIndexOf(stoken, fpos);
        if (spos<0) return false;

        //check for the occurrence of the end token, starting immediately _after_
        //the starting token. If it doesn't occur, -or- if it occurs _before_ the
        //fpos, then return false. Otherwise, it must be embedded.
        int epos = str.indexOf(etoken, spos+stoken.length());
        if (epos<0 || epos<fpos) return false;
        else return true;
    }

    /**
     * In order to use the Javac taskdef from inside our localize taskdef,
     * we have to be able to set the project name
     */
    class InternalJavac extends Javac {

        public InternalJavac() {
//            this.setTaskName("localize");
            this.setTaskName(Localize.this.getTaskName());
        }

        public void setProject(Project iproject) {
            this.project = iproject;
        }
    }

    public static void main(String args[]) {
        Localize loc = new Localize();

        //this one shouldn't get replaced
        System.out.println(loc.qualifiedStrReplace("<select id=\"AbstractBComponent_DebugLevel\" name=\"D1\" size=\"1\">", "Debug", "zipper"));

        //this one should
        System.out.println(loc.qualifiedStrReplace("<select id=\"Abstract\" blah=\"DebugLevel\" name=\"D1\" size=\"1\">", "Debug", "zipper"));
    }
}

